/**********************************************************************
*                                                                     *
*    MODULE-NAME = fftwf.c                                            *
*                                                                     *
*    DESCRIPTIVE-NAME = FFTW C Interfaces (single precision).         *
*                                                                     *
*    Licensed Materials - Property of IBM                             *
*                                                                     *
*    5765-F82                                                         *
*    5765-G17                                                         *
*    5765-H25                                                         *
*    5765-L51                                                         *
*                                                                     *
*    Copyright IBM Corp. 2008, 2013                                   *
*    Copyright (c) 2003, 2007 Massachusetts Institute of Technology   *
*    Copyright (c) 2003, 2007 Matteo Frigo                            *
*    All Rights Reserved                                              *
*                                                                     *
*    US Government Users Restricted Rights - Use, duplication or      *
*    disclosure restricted by GSA ADP Schedule Contract with IBM Corp.*
*                                                                     *
*    FUNCTION = Call FFTW functions from C                            *
*               for single precision computation.                     *
*                                                                     *
*    NOTES =                                                          *
*                                                                     *
*      DEPENDENCIES = (1) fftw3_essl.h                                *
*                     (2) essl.h                                      *
*                                                                     *
*      RESTRICTIONS = None                                            *
*                                                                     *
*      PROCESSOR = C/C++ Compiler                                     *
*                                                                     *
***********************************************************************
*                                                                     *
*      PURPOSE = SEE FUNCTION                                         *
*                                                                     *
*      LINKAGE = NONE                                                 *
*                                                                     *
*      CONTROL BLOCKS = NONE                                          *
*                                                                     *
*      TABLES = NONE                                                  *
*                                                                     *
*      MACROS = NONE                                                  *
*                                                                     *
*      CHANGE-ACTIVITY = NONE                                         *
*                                                                     *
***********************************************************************
*                ALGORITHM DESCRIPTION                                *
*                                                                     *
* REFERENCES  = NONE                                                  *
*                                                                     *
* FUNCTION    =   See Above                                           *
*                                                                     *
***********************************************************************
*                    ALGORITHM EXAMPLES                               *
*                                                                     *
*   NONE                                                              *
*                                                                     *
***********************************************************************
*                                                                     *
*                PROCEDURAL LOGIC SECTION                             *
*                                                                     *
*    1. To compute complex FFTs scftd is called.                      *
*       For complex-to-real and real-to-complex FFTs we call scrftd   *
*       and srcftd, respectively.                                     *
*                                                                     *
*    2. All errors are handled inside of the routines scftd, scrftd,  *
*       and srcftd.                                                   *
*                                                                     *
*    3. To perform the complex-to-real FFTs the auxiliary function    *
*       fftwf_dft_c2r_x is used. This function uses the input array   *
*       as a temporary storage in the multidimensional transforms.    *
*       Since the algorithm in fftwf_dft_c2r_x is not input-preserving*
*       the planners for multidimensional c2r transforms return NULL  *
*       by passing the flag FFTW_PRESERVE_INPUT.                      *
*                                                                     *
*    4. All FFTW plan functions return the NULL pointer if we are     *
*       unable to allocate memory. In all other situations, they      *
*       return a pointer on the structure fftwf_esv_plan_s which      *
*       contains all needed data for scftd (scrftd, srcftd).          *
*                                                                     *
*    5. The one dimensional FFTs of the size one are handled by the   *
*       auxiliary functions fftwf_dft_size_one,                       *
*       fftwf_dft_r2c_size_one, and fftwf_dft_c2r_size_one.           *
*                                                                     *
**********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#define __ESVERR
#include <essl.h>
#include "fftw3_essl.h"
extern int enotrm();
extern void errset(int*, int*, int*, int*, int(**)(), int*);
extern void errsav(int*, void *);
extern void errstr(int*, void *);
extern int posix_memalign(void **, size_t, size_t);

#ifdef _ESSL_64
    #define _ESVINT64 long
#else
    #define _ESVINT64 _ESVINT
#endif

struct fftwf_esv_plan_s
{
  int     type;
  int     d;
  void*   x;
  int*    incx;
  int*    incmx;
  void*   y;
  int*    incy;
  int*    incmy;
  int*    n;
  int*    m;
  int*    isign;
  float   scale;
  double* aux1;
  int     naux1;
  double* aux2;
  int     naux2;
  int     howmany_rank;
  int*    mm;
};

void fftwf_dft_c2r_x(int,int,void*,int*,int,void*,int*,int,int*,int, 
		    int*,float,double*,int*,double*,int*);

void fftwf_dft_size_one(fftwf_complex *x, fftwf_complex *y, 
		       int m, int incmx, int incmy);

void fftwf_dft_r2c_size_one(float *x, fftwf_complex *y, 
			   int m, int incmx, int incmy);

void fftwf_dft_c2r_size_one(fftwf_complex *x, float *y, 
			   int m, int incmx, int incmy);

#define ESV_C2C (1)
#define ESV_C2R (2)
#define ESV_R2C (3)


/* C2C ----------------------------------------------------------- */

fftwf_plan fftwf_plan_dft_1d(int n, fftwf_complex *in, fftwf_complex *out, 
			   int sign,unsigned flags)
{
  fftwf_plan p;
  double *aux1d, *aux2d;
  int (*iusadr) (); 
  int ierno,inoal,inomes,itrace,irange,dummy;
  char s2015[8];
  int init;
  init = 1;
  if ((flags & FFTW_NO_SIMD) == FFTW_NO_SIMD) init = 2;
  if ((flags & FFTW_UNALIGNED) == FFTW_UNALIGNED) init = 2;
  /* allocate the structure fftwf_esv_plan_s                  */
  p = (fftwf_plan) malloc(sizeof(struct fftwf_esv_plan_s)); 
  if (p == NULL) return NULL;
  /* fill the structure fftwf_esv_plan_s                      */
  p->howmany_rank = 1;
  p->mm = NULL;
  /* type of the transform: complex-to-complex,
     complex-to-real, or real-to-complex                      */
  p->type = ESV_C2C;
  /* dimension of the transform                               */
  p->d = 1;
  /* input array                                              */
  p->x = in;
  /* output array                                             */
  p->y = out;
  /* scale factor                                             */
  p->scale = 1.0;
  /* allocate all the integer arrays at the same time         */
  p->incx = (int *) malloc(7*sizeof(int));
  if (p->incx == NULL)
    {
      free(p);
      return NULL;
    }
  /* strides for the input array                              */
  p->incx[0] = 1;
  p->incmx = &(p->incx[1]);
  p->incmx[0] = n;
  /* strides for the output array                             */
  p->incy = &(p->incmx[1]);
  p->incy[0] = 1;
  p->incmy = &(p->incy[1]);
  p->incmy[0] = n;
  /* length of the transform                                  */
  p->n = &(p->incmy[1]);
  p->n[0] = n;
  /* number of arrays to be transformed                       */
  p->m = &(p->n[1]);
  p->m[0] = 1;
  /* direction of the transform                               */
  p->isign = &(p->m[1]);
  p->isign[0] = -sign;
  /* size one dft                                             */
  if (n == 1)
    {
      p->d = 0;
      return p;
    }
  /* determine size of aux1 and aux2 arrays                   */
  einfo(0,&dummy,&dummy);
  iusadr = enotrm;
  ierno = 2015;
  inoal = 0;
  inomes = -1;
  itrace = 0;
  irange = 2015;
  errsav(&ierno,s2015);
  errset(&ierno,&inoal,&inomes,&itrace,&iusadr,&irange);
  p->naux1 = 7*(p->d + 1) + 1;
  p->naux2 = 0;
  aux1d = (double *) fftwf_malloc((_ESVINT64)p->naux1*sizeof(double));
  aux2d = (double *) fftwf_malloc((_ESVINT64)2*sizeof(double));
  if ((aux1d == NULL) || (aux2d == NULL))
    {
      if (p->incx != NULL) free(p->incx);
      if (aux1d != NULL) fftwf_free(aux1d);
      if (aux2d != NULL) fftwf_free(aux2d);
      free(p);
      return NULL;
    }
  scftd(init,p->d,p->x,p->incx,p->incmx[0],
        p->y,p->incy,p->incmy[0],
        p->n,p->m[0],p->isign,p->scale,
        aux1d,&p->naux1,aux2d,&p->naux2);
  fftwf_free(aux1d);
  fftwf_free(aux2d);
  errstr(&ierno,s2015);
  /* allocate aux1 and aux2 arrays                            */
  p->aux1 = (double *) fftwf_malloc((_ESVINT64)p->naux1*sizeof(double));
  p->aux2 = (double *) fftwf_malloc((_ESVINT64)p->naux2*sizeof(double));
  if ((p->aux1 == NULL) || (p->aux2 == NULL))
    {
      if (p->incx != NULL) free(p->incx);
      if (p->aux1 != NULL) fftwf_free(p->aux1);
      if (p->aux2 != NULL) fftwf_free(p->aux2);
      free(p);
      return NULL;
    }
  /* initialization of scftd                                   */
  scftd(init,p->d,p->x,p->incx,p->incmx[0],
        p->y,p->incy,p->incmy[0],
        p->n,p->m[0],p->isign,p->scale,
        p->aux1,&p->naux1,p->aux2,&p->naux2);
  return p; 
}

fftwf_plan fftwf_plan_dft_2d(int nx, int ny, 
			   fftwf_complex *in, fftwf_complex *out, 
			   int sign,unsigned flags)
{
  fftwf_plan p;
  double *aux1d, *aux2d;
  int (*iusadr) (); 
  int ierno,inoal,inomes,itrace,irange,dummy;
  char s2015[8];
  int d, i;
  int init;
  init = 1;
  if ((flags & FFTW_NO_SIMD) == FFTW_NO_SIMD) init = 2;
  if ((flags & FFTW_UNALIGNED) == FFTW_UNALIGNED) init = 2;
  /* allocate the structure fftwf_esv_plan_s                  */
  p = (fftwf_plan) malloc(sizeof(struct fftwf_esv_plan_s)); 
  if (p == NULL) return NULL;
  /* fill the structure fftwf_esv_plan_s                      */
  p->howmany_rank = 1;
  p->mm = NULL;
  /* type of the transform: complex-to-complex,
     complex-to-real, or real-to-complex                      */
  p->type = ESV_C2C;
  /* dimension of the transform                               */
  p->d = 2;
  /* input array                                              */
  p->x = in;
  /* output array                                             */
  p->y = out;
  /* scale factor                                             */
  p->scale = 1.0;
  /* allocate all the integer arrays at the same time         */
  p->incx = (int *) malloc(11*sizeof(int));
  if (p->incx == NULL)
    {
      free(p);
      return NULL;
    }
  /* strides for the input array                              */
  p->incx[0] = 1;
  p->incx[1] = ny;
  p->incmx = &(p->incx[2]);
  p->incmx[0] = nx*ny;
  /* strides for the output array                             */
  p->incy = &(p->incmx[1]);
  p->incy[0] = 1;
  p->incy[1] = ny;
  p->incmy = &(p->incy[2]);
  p->incmy[0] = nx*ny;
  /* length of the transform                                  */
  p->n = &(p->incmy[1]);
  p->n[0] = ny;
  p->n[1] = nx;
  /* number of arrays to be transformed                       */
  p->m = &(p->n[2]);
  p->m[0] = 1;
  /* direction of the transform                               */
  p->isign = &(p->m[1]);
  p->isign[0] = -sign;
  p->isign[1] = -sign;
  /* size one dft                                             */
  d = 0;
  for (i=0; i<2; i++)
    {
      if (p->n[i] != 1)
	{
	  if (d < i)
	    {
	      p->n[d] = p->n[i];
	      p->incx[d] = p->incx[i];
	      p->incy[d] = p->incy[i];
	      p->isign[d] = p->isign[i];
	    }
	  d = d+1;
	}
    }
  p->d = d;
  if (d == 0) return p;
  /* determine size of aux1 and aux2 arrays                   */
  einfo(0,&dummy,&dummy);
  iusadr = enotrm;
  ierno = 2015;
  inoal = 0;
  inomes = -1;
  itrace = 0;
  irange = 2015;
  errsav(&ierno,s2015);
  errset(&ierno,&inoal,&inomes,&itrace,&iusadr,&irange);
  p->naux1 = 7*(p->d + 1) + 1;
  p->naux2 = 0;
  aux1d = (double *) fftwf_malloc((_ESVINT64)p->naux1*sizeof(double));
  aux2d = (double *) fftwf_malloc((_ESVINT64)2*sizeof(double));
  if ((aux1d == NULL) || (aux2d == NULL))
    {
      if (p->incx != NULL) free(p->incx);
      if (aux1d != NULL) fftwf_free(aux1d);
      if (aux2d != NULL) fftwf_free(aux2d);
      free(p);
      return NULL;
    }
  scftd(init,p->d,p->x,p->incx,p->incmx[0],
        p->y,p->incy,p->incmy[0],
        p->n,p->m[0],p->isign,p->scale,
        aux1d,&p->naux1,aux2d,&p->naux2);
  fftwf_free(aux1d);
  fftwf_free(aux2d);
  errstr(&ierno,s2015);
  /* allocate aux1 and aux2 arrays                            */
  p->aux1 = (double *) fftwf_malloc((_ESVINT64)p->naux1*sizeof(double));
  p->aux2 = (double *) fftwf_malloc((_ESVINT64)p->naux2*sizeof(double));
  if ((p->aux1 == NULL) || (p->aux2 == NULL))
    {
      if (p->incx != NULL) free(p->incx);
      if (p->aux1 != NULL) fftwf_free(p->aux1);
      if (p->aux2 != NULL) fftwf_free(p->aux2);
      free(p);
      return NULL;
    }
  /* initialization of scftd                                  */
  scftd(init,p->d,p->x,p->incx,p->incmx[0],
        p->y,p->incy,p->incmy[0],
        p->n,p->m[0],p->isign,p->scale,
        p->aux1,&p->naux1,p->aux2,&p->naux2);

  return p; 
}

fftwf_plan fftwf_plan_dft_3d(int nx, int ny, int nz, 
			   fftwf_complex *in, fftwf_complex *out, 
			   int sign,unsigned flags)
{
  fftwf_plan p;
  double *aux1d, *aux2d;
  int (*iusadr) (); 
  int ierno,inoal,inomes,itrace,irange,dummy;
  char s2015[8];
  int d, i;
  int init;
  init = 1;
  if ((flags & FFTW_NO_SIMD) == FFTW_NO_SIMD) init = 2;
  if ((flags & FFTW_UNALIGNED) == FFTW_UNALIGNED) init = 2;
  /* allocate the structure fftwf_esv_plan_s                   */
  p = (fftwf_plan) malloc(sizeof(struct fftwf_esv_plan_s)); 
  if (p == NULL) return NULL;
  /* fill the structure fftwf_esv_plan_s                       */
  p->howmany_rank = 1;
  p->mm = NULL;
  /* type of the transform: complex-to-complex,
     complex-to-real, or real-to-complex                      */
  p->type = ESV_C2C;
  /* dimension of the transform                               */
  p->d = 3;
  /* input array                                              */
  p->x = in;
  /* output array                                             */
  p->y = out;
  /* scale factor                                             */
  p->scale = 1.0;
  /* allocate all the integer arrays at the same time         */
  p->incx = (int *) malloc(15*sizeof(int));
  if (p->incx == NULL)
    {
      free(p);
      return NULL;
    }
  /* strides for the input array                              */
  p->incx[0] = 1;
  p->incx[1] = nz;
  p->incx[2] = ny*nz;
  p->incmx = &(p->incx[3]);
  p->incmx[0] = nx*ny*nz;
  /* strides for the output array                             */
  p->incy = &(p->incmx[1]);
  p->incy[0] = 1;
  p->incy[1] = nz;
  p->incy[2] = ny*nz;
  p->incmy = &(p->incy[3]);
  p->incmy[0] = nx*ny*nz;
  /* length of the transform                                  */
  p->n = &(p->incmy[1]);
  p->n[0] = nz;
  p->n[1] = ny;
  p->n[2] = nx;
  /* number of arrays to be transformed                       */
  p->m = &(p->n[3]);
  p->m[0] = 1;
  /* direction of the transform                               */
  p->isign = &(p->m[1]);
  p->isign[0] = -sign;
  p->isign[1] = -sign;
  p->isign[2] = -sign;
  /* size one dft                                             */
  d = 0;
  for (i=0; i<3; i++)
    {
      if (p->n[i] != 1)
	{
	  if (d < i)
	    {
	      p->n[d] = p->n[i];
	      p->incx[d] = p->incx[i];
	      p->incy[d] = p->incy[i];
	      p->isign[d] = p->isign[i];
	    }
	  d = d+1;
	}
    }
  p->d = d;
  if (d == 0) return p;
  /* determine size of aux1 and aux2 arrays                   */
  einfo(0,&dummy,&dummy);
  iusadr = enotrm;
  ierno = 2015;
  inoal = 0;
  inomes = -1;
  itrace = 0;
  irange = 2015;
  errsav(&ierno,s2015);
  errset(&ierno,&inoal,&inomes,&itrace,&iusadr,&irange);
  p->naux1 = 7*(p->d + 1) + 1;
  p->naux2 = 0;
  aux1d = (double *) fftwf_malloc((_ESVINT64)p->naux1*sizeof(double));
  aux2d = (double *) fftwf_malloc((_ESVINT64)2*sizeof(double));
  if ((aux1d == NULL) || (aux2d == NULL))
    {
      if (p->incx != NULL) free(p->incx);
      if (aux1d != NULL) fftwf_free(aux1d);
      if (aux2d != NULL) fftwf_free(aux2d);
      free(p);
      return NULL;
    }
  scftd(init,p->d,p->x,p->incx,p->incmx[0],
        p->y,p->incy,p->incmy[0],
        p->n,p->m[0],p->isign,p->scale,
        aux1d,&p->naux1,aux2d,&p->naux2);
  fftwf_free(aux1d);
  fftwf_free(aux2d);
  errstr(&ierno,s2015);
  /* allocate aux1 and aux2 arrays                            */
  p->aux1 = (double *) fftwf_malloc((_ESVINT64)p->naux1*sizeof(double));
  p->aux2 = (double *) fftwf_malloc((_ESVINT64)p->naux2*sizeof(double));
  if ((p->aux1 == NULL) || (p->aux2 == NULL))
    {
      if (p->incx != NULL) free(p->incx);
      if (p->aux1 != NULL) fftwf_free(p->aux1);
      if (p->aux2 != NULL) fftwf_free(p->aux2);
      free(p);
      return NULL;
    }
  /* initialization of scftd                                   */
  scftd(init,p->d,p->x,p->incx,p->incmx[0],
        p->y,p->incy,p->incmy[0],
        p->n,p->m[0],p->isign,p->scale,
        p->aux1,&p->naux1,p->aux2,&p->naux2);

  return p; 
}

fftwf_plan fftwf_plan_dft(int rank, const int *n, 
			fftwf_complex *in, fftwf_complex *out, 
			int sign,unsigned flags)
{
  fftwf_plan p;
  double *aux1d, *aux2d;
  int (*iusadr) (); 
  int ierno,inoal,inomes,itrace,irange,dummy;
  char s2015[8];
  int d, i;
  int init;
  init = 1;
  if ((flags & FFTW_NO_SIMD) == FFTW_NO_SIMD) init = 2;
  if ((flags & FFTW_UNALIGNED) == FFTW_UNALIGNED) init = 2;
  /* allocate the structure fftwf_esv_plan_s                  */
  p = (fftwf_plan) malloc(sizeof(struct fftwf_esv_plan_s)); 
  if (p == NULL) return NULL;
  /* fill the structure fftwf_esv_plan_s                      */
  p->howmany_rank = 1;
  p->mm = NULL;
  /* type of the transform: complex-to-complex,
     complex-to-real, or real-to-complex                      */
  p->type = ESV_C2C;
  /* dimension of the transform                               */
  p->d = rank;
  /* input array                                              */
  p->x = in;
  /* output array                                             */
  p->y = out;
  /* scale factor                                             */
  p->scale = 1.0;
  /* allocate all the integer arrays at the same time         */
  p->incx = (int *) malloc((_ESVINT64)(4*rank+3)*sizeof(int));  
  if (p->incx == NULL)
    {
      free(p);
      return NULL;
    }
  /* handle rank = 0                                          */
  if (rank == 0)
    {
      p->incmx = p->incx;      
      p->incmy = &(p->incmx[1]);    
      p->m = &(p->incmy[1]); 

      p->incmx[0] = 1;
      p->incmy[0] = 1;
      p->m[0] = 1;

      return p;
    }
  /* strides for the input array                              */
  /* strides for the output array                             */
  /* length of the transform                                  */
  /* number of arrays to be transformed                       */
  /* direction of the transform                               */
  p->incmx = &(p->incx[rank]);
  p->incy = &(p->incmx[1]);
  p->incmy = &(p->incy[rank]);
  p->n = &(p->incmy[1]);
  p->m = &(p->n[rank]);
  p->isign = &(p->m[1]);

  p->incx[0] = 1;
  p->incy[0] = 1;
  p->n[0] = n[rank-1];
  p->isign[0] = -sign;
  for (i=1; i<rank; i++)
    {
      p->incx[i] = n[rank-i]*p->incx[i-1];
      p->incy[i] = p->incx[i];
      p->n[i] = n[rank-1-i];
      p->isign[i] = -sign;
    }
  p->incmx[0] = n[0]*p->incx[rank-1];
  p->incmy[0] = p->incmx[0];
  p->m[0] = 1;
  /* determine size of aux1 and aux2 arrays                   */
  einfo(0,&dummy,&dummy);
  iusadr = enotrm;
  ierno = 2015;
  inoal = 0;
  inomes = -1;
  itrace = 0;
  irange = 2015;
  errsav(&ierno,s2015);
  errset(&ierno,&inoal,&inomes,&itrace,&iusadr,&irange);
  p->naux1 = 7*(p->d + 1) + 1;
  p->naux2 = 0;
  aux1d = (double *) fftwf_malloc((_ESVINT64)p->naux1*sizeof(double));
  aux2d = (double *) fftwf_malloc((_ESVINT64)2*sizeof(double));
  if ((aux1d == NULL) || (aux2d == NULL))
    {
      if (p->incx != NULL) free(p->incx);
      if (aux1d != NULL) fftwf_free(aux1d);
      if (aux2d != NULL) fftwf_free(aux2d);
      free(p);
      return NULL;
    }
  scftd(init,p->d,p->x,p->incx,p->incmx[0],
        p->y,p->incy,p->incmy[0],
        p->n,p->m[0],p->isign,p->scale,
        aux1d,&p->naux1,aux2d,&p->naux2);
  fftwf_free(aux1d);
  fftwf_free(aux2d);
  errstr(&ierno,s2015);
  /* allocate aux1 and aux2 arrays                            */
  p->aux1 = (double *) fftwf_malloc((_ESVINT64)p->naux1*sizeof(double));
  p->aux2 = (double *) fftwf_malloc((_ESVINT64)p->naux2*sizeof(double));
  if ((p->aux1 == NULL) || (p->aux2 == NULL && p->naux2 != 0))
    {
      if (p->incx != NULL) free(p->incx);
      if (p->aux1 != NULL) fftwf_free(p->aux1);
      if (p->aux2 != NULL) fftwf_free(p->aux2);
      free(p);
      return NULL;
    }
  /* initialization of scftd                                  */
  scftd(init,p->d,p->x,p->incx,p->incmx[0],
        p->y,p->incy,p->incmy[0],
        p->n,p->m[0],p->isign,p->scale,
        p->aux1,&p->naux1,p->aux2,&p->naux2);

  return p; 
}

fftwf_plan fftwf_plan_many_dft(int rank, const int *n, int howmany,
			     fftwf_complex *in, const int *inembed,
			     int istride, int idist,
			     fftwf_complex *out, const int *onembed,
			     int ostride, int odist,  
			     int sign,unsigned flags)
{
  fftwf_plan p;
  double *aux1d, *aux2d;
  int (*iusadr) (); 
  int ierno,inoal,inomes,itrace,irange,dummy;
  char s2015[8];
  int d, i;
  int init;
  init = 1;
  if ((flags & FFTW_NO_SIMD) == FFTW_NO_SIMD) init = 2;
  if ((flags & FFTW_UNALIGNED) == FFTW_UNALIGNED) init = 2;
  /* allocate the structure fftwf_esv_plan_s                  */
  p = (fftwf_plan) malloc(sizeof(struct fftwf_esv_plan_s)); 
  if (p == NULL) return NULL;
  /* fill the structure fftwf_esv_plan_s                      */
  p->howmany_rank = 1;
  p->mm = NULL;
  /* type of the transform: complex-to-complex,
     complex-to-real, or real-to-complex                      */
  p->type = ESV_C2C;
  /* dimension of the transform                               */
  p->d = rank;
  /* input array                                              */
  p->x = in;
  /* output array                                             */
  p->y = out;
  /* scale factor                                             */
  p->scale = 1.0;
  /* allocate all the integer arrays at the same time         */
  p->incx = (int *) malloc((_ESVINT64)(4*rank+3)*sizeof(int));
  if (p->incx == NULL)
    {
      free(p);
      return NULL;
    }
  /* handle rank = 0                                          */
  if (rank == 0)
    {
      p->incmx = p->incx;      
      p->incmy = &(p->incmx[1]);    
      p->m = &(p->incmy[1]); 

      p->incmx[0] = idist;
      p->incmy[0] = odist;
      p->m[0] = howmany;

      return p;
    }
  /* strides for the input array                              */
  if (inembed == NULL)
    {
      p->incx[0] = istride;
      for (i=1; i<rank; i++)
	{
	  p->incx[i] = n[rank-i]*p->incx[i-1];
	}
    }
  else
    {
      p->incx[0] = istride;
      for (i=1; i<rank; i++)
	{
	  p->incx[i] = inembed[rank-i]*p->incx[i-1];
	}
    }
  p->incmx = &(p->incx[rank]);
  p->incmx[0] = idist;
  /* strides for the output array                             */
  p->incy = &(p->incmx[1]);
  if (onembed == NULL)
    {
      p->incy[0] = ostride;
      for (i=1; i<rank; i++)
	{
	  p->incy[i] = n[rank-i]*p->incy[i-1];
	}
    }
  else
    {
      p->incy[0] = ostride;
      for (i=1; i<rank; i++)
	{
	  p->incy[i] = onembed[rank-i]*p->incy[i-1];
	}
    }
  p->incmy = &(p->incy[rank]);
  p->incmy[0] = odist;
  /* length of the transform                                  */
  p->n = &(p->incmy[1]);
  for (i=0; i<rank; i++)
    {
      p->n[i] = n[rank-1-i];
    }
  /* number of arrays to be transformed                       */
  p->m = &(p->n[rank]);
  p->m[0] = howmany;
  /* direction of the transform                               */
  p->isign = &(p->m[1]);
  for (i=0; i<rank; i++)
    {
      p->isign[i] = -sign;
    }
  /* determine size of aux1 and aux2 arrays                   */
  einfo(0,&dummy,&dummy);
  iusadr = enotrm;
  ierno = 2015;
  inoal = 0;
  inomes = -1;
  itrace = 0;
  irange = 2015;
  errsav(&ierno,s2015);
  errset(&ierno,&inoal,&inomes,&itrace,&iusadr,&irange);
  p->naux1 = 7*(p->d + 1) + 1;
  p->naux2 = 0;
  aux1d = (double *) fftwf_malloc((_ESVINT64)p->naux1*sizeof(double));
  aux2d = (double *) fftwf_malloc((_ESVINT64)2*sizeof(double));
  if ((aux1d == NULL) || (aux2d == NULL))
    {
      if (p->incx != NULL) free(p->incx);
      if (aux1d != NULL) fftwf_free(aux1d);
      if (aux2d != NULL) fftwf_free(aux2d);
      free(p);
      return NULL;
    }
  scftd(init,p->d,p->x,p->incx,p->incmx[0],
        p->y,p->incy,p->incmy[0],
        p->n,p->m[0],p->isign,p->scale,
        aux1d,&p->naux1,aux2d,&p->naux2);
  fftwf_free(aux1d);
  fftwf_free(aux2d);
  errstr(&ierno,s2015);
  /* allocate aux1 and aux2 arrays                            */
  p->aux1 = (double *) fftwf_malloc((_ESVINT64)p->naux1*sizeof(double));
  p->aux2 = (double *) fftwf_malloc((_ESVINT64)p->naux2*sizeof(double));
  if ((p->aux1 == NULL) || (p->aux2 == NULL && p->naux2 != 0))
    {
      if (p->incx != NULL) free(p->incx);
      if (p->aux1 != NULL) fftwf_free(p->aux1);
      if (p->aux2 != NULL) fftwf_free(p->aux2);
      free(p);
      return NULL;
    }
  /* initialization of scftd                                  */
  scftd(init,p->d,p->x,p->incx,p->incmx[0],
        p->y,p->incy,p->incmy[0],
        p->n,p->m[0],p->isign,p->scale,
        p->aux1,&p->naux1,p->aux2,&p->naux2);

  return p; 
}

fftwf_plan fftwf_plan_guru_dft(int rank, const fftwf_iodim *dims,
			     int howmany_rank, const fftwf_iodim *howmany_dims,
			     fftwf_complex *in, fftwf_complex *out,
			     int sign,unsigned flags)
{
  fftwf_plan p;
  double *aux1d, *aux2d;
  int (*iusadr) (); 
  int ierno,inoal,inomes,itrace,irange,dummy;
  char s2015[8];
  int size;
  int d, i;
  int init, stride;
  init = 1;
  if ((flags & FFTW_NO_SIMD) == FFTW_NO_SIMD) init = 2;
  if ((flags & FFTW_UNALIGNED) == FFTW_UNALIGNED) init = 2;
  /* allocate the structure fftwf_esv_plan_s                  */
  p = (fftwf_plan) malloc(sizeof(struct fftwf_esv_plan_s)); 
  if (p == NULL) return NULL;
  /* fill the structure fftwf_esv_plan_s                      */
  /* type of the transform: complex-to-complex,
     complex-to-real, or real-to-complex                      */
  p->type = ESV_C2C;
  /* dimension of the transform                               */
  p->d = rank;
  /* input array                                              */
  p->x = in;
  /* output array                                             */
  p->y = out;
  /* scale factor                                             */
  p->scale = 1.0;
  /* allocate all the integer arrays at the same time         */
  if (howmany_rank > 1)
    {
      size = 4*(rank + howmany_rank);
    }
  else
    {
      size = 4*rank + 3;
    }
  p->incx = (int *) malloc((_ESVINT64)size*sizeof(int));
  if (p->incx == NULL)
    {
      free(p);
      return NULL;
    }
  /* strides for the input array                              */
  /* strides for the output array                             */
  /* length of the transform                                  */
  /* direction of the transform                               */
  p->incy = &(p->incx[rank]);
  p->n = &(p->incy[rank]);
  p->isign = &(p->n[rank]);
  for (i=0; i<rank; i++)
    {
      p->incx[i] = dims[rank-1-i].is;
      p->incy[i] = dims[rank-1-i].os;
      p->n[i] = dims[rank-1-i].n;
      p->isign[i] = -sign;
    }
  /* p->incmx, p->incmy, p->m, p->howmany_rank, and p->mm
     depend on the parameter howmany_rank                     */
  if (howmany_rank == 0)
    {
      p->howmany_rank = 1;
      p->mm = NULL;
      p->m = &(p->isign[rank]);
      p->incmx = &(p->m[1]);
      p->incmy = &(p->incmx[1]);

      if (rank == 0)
	{
	  p->m[0] = 1;
	  p->incmx[0] = 1;
	  p->incmy[0] = 1;

	}
      else
	{
	  p->m[0] = 1;
	  p->incmx[0] = p->n[rank-1]*p->incx[rank-1];
	  p->incmy[0] = p->n[rank-1]*p->incy[rank-1];
	}
    }
  else if (howmany_rank == 1)
    {
      p->howmany_rank = 1;
      p->mm = NULL;
      p->m = &(p->isign[rank]);
      p->incmx = &(p->m[1]);
      p->incmy = &(p->incmx[1]);
 
      p->m[0] = howmany_dims[0].n;
      p->incmx[0] = howmany_dims[0].is;
      p->incmy[0] = howmany_dims[0].os;
    }
  else
    {
      p->howmany_rank = howmany_rank;
      p->m = &(p->isign[rank]);
      p->incmx = &(p->m[howmany_rank]);
      p->incmy = &(p->incmx[howmany_rank]);
      p->mm = &(p->incmy[howmany_rank]); 

      p->m[0] = howmany_dims[0].n;
      p->incmx[0] = howmany_dims[0].is;
      p->incmy[0] = howmany_dims[0].os;
      p->mm[0] = 1;
      for (i=1; i<howmany_rank; i++)
	{
	  p->m[i] = howmany_dims[i].n;
	  p->incmx[i] = howmany_dims[i].is;
	  p->incmy[i] = howmany_dims[i].os;

	  p->mm[i] = p->mm[0];
	  p->mm[0] = p->mm[0]*p->m[i];
	}
      /* here also we have to decide whether we will use      */
      /* SIMD performance enhancements                        */
      /* (check whether p->incmx[i] and p->incmy[i] are       */
      /*  even for single precision computation)              */
      for (i=1; i<howmany_rank; i++)
	{
	  stride = p->incmx[i]/2;
	  if (p->incmx[i] != 2*stride) init = 2;
	  stride = p->incmy[i]/2;
	  if (p->incmy[i] != 2*stride) init = 2;
	}
    }

  /* handle rank = 0                                          */
  if (p->d == 0) return p;

  /* determine size of aux1 and aux2 arrays                   */
  einfo(0,&dummy,&dummy);
  iusadr = enotrm;
  ierno = 2015;
  inoal = 0;
  inomes = -1;
  itrace = 0;
  irange = 2015;
  errsav(&ierno,s2015);
  errset(&ierno,&inoal,&inomes,&itrace,&iusadr,&irange);
  p->naux1 = 7*(p->d + 1) + 1;
  p->naux2 = 0;
  aux1d = (double *) fftwf_malloc((_ESVINT64)p->naux1*sizeof(double));
  aux2d = (double *) fftwf_malloc((_ESVINT64)2*sizeof(double));
  if ((aux1d == NULL) || (aux2d == NULL))
    {
      if (p->incx != NULL) free(p->incx);
      if (aux1d != NULL) fftwf_free(aux1d);
      if (aux2d != NULL) fftwf_free(aux2d);
      free(p);
      return NULL;
    }
  scftd(init,p->d,p->x,p->incx,p->incmx[0],
        p->y,p->incy,p->incmy[0],
        p->n,p->m[0],p->isign,p->scale,
        aux1d,&p->naux1,aux2d,&p->naux2);
  fftwf_free(aux1d);
  fftwf_free(aux2d);
  errstr(&ierno,s2015);
  /* allocate aux1 and aux2 arrays                            */
  p->aux1 = (double *) fftwf_malloc((_ESVINT64)p->naux1*sizeof(double));
  p->aux2 = (double *) fftwf_malloc((_ESVINT64)p->naux2*sizeof(double));
  if ((p->aux1 == NULL) || (p->aux2 == NULL && p->naux2 != 0))
    {
      if (p->incx != NULL) free(p->incx);
      if (p->aux1 != NULL) fftwf_free(p->aux1);
      if (p->aux2 != NULL) fftwf_free(p->aux2);
      free(p);
      return NULL;
    }
  /* initialization of scftd                                  */
  scftd(init,p->d,p->x,p->incx,p->incmx[0],
        p->y,p->incy,p->incmy[0],
        p->n,p->m[0],p->isign,p->scale,
        p->aux1,&p->naux1,p->aux2,&p->naux2);

  return p; 
}

/* R2C ----------------------------------------------------------- */


fftwf_plan fftwf_plan_dft_r2c_1d(int n, float *in, 
			       fftwf_complex *out, unsigned flags)
{
  fftwf_plan p;
  double *aux1d, *aux2d;
  int (*iusadr) (); 
  int ierno,inoal,inomes,itrace,irange,dummy;
  char s2015[8];
  int init;
  init = 1;
  if ((flags & FFTW_NO_SIMD) == FFTW_NO_SIMD) init = 2;
  if ((flags & FFTW_UNALIGNED) == FFTW_UNALIGNED) init = 2;
  /* allocate the structure fftwf_esv_plan_s                  */
  p = (fftwf_plan) malloc(sizeof(struct fftwf_esv_plan_s)); 
  if (p == NULL) return NULL;
  /* fill the structure fftwf_esv_plan_s                      */
  p->howmany_rank = 1;
  p->mm = NULL;
  /* type of the transform: complex-to-complex,
     complex-to-real, or real-to-complex                      */
  p->type = ESV_R2C;
  /* dimension of the transform                               */
  p->d = 1;
  /* input array                                              */
  p->x = in;
  /* output array                                             */
  p->y = out;
  /* scale factor                                             */
  p->scale = 1.0;
  /* allocate all the integer arrays at the same time         */
  p->incx = (int *) malloc(7*sizeof(int));
  if (p->incx == NULL)
    {
      free(p);
      return NULL;
    }
  /* strides for the input array                              */
  p->incx[0] = 1;
  p->incmx = &(p->incx[1]);
  if ((void*) in == (void*) out)
    {
      p->incmx[0] = 2*(n/2+1);      
    }
  else
    {
      p->incmx[0] = n;
    }
  /* strides for the output array                             */
  p->incy = &(p->incmx[1]);
  p->incy[0] = 1;
  p->incmy = &(p->incy[1]);
  p->incmy[0] = n/2+1;
  /* length of the transform                                  */
  p->n = &(p->incmy[1]);
  p->n[0] = n;
  /* number of arrays to be transformed                       */
  p->m = &(p->n[1]);
  p->m[0] = 1;
  /* direction of the transform                               */
  p->isign = &(p->m[1]);
  p->isign[0] = 1;
  /* size one dft                                             */
  if (n == 1)
    {
      p->d = 0;
      return p;
    }
  /* determine size of aux1 and aux2 arrays                   */
  einfo(0,&dummy,&dummy);
  iusadr = enotrm;
  ierno = 2015;
  inoal = 0;
  inomes = -1;
  itrace = 0;
  irange = 2015;
  errsav(&ierno,s2015);
  errset(&ierno,&inoal,&inomes,&itrace,&iusadr,&irange);
  p->naux1 = 15;
  p->naux2 = 0;
  aux1d = (double *) fftwf_malloc((_ESVINT64)p->naux1*sizeof(double));
  aux2d = (double *) fftwf_malloc((_ESVINT64)2*sizeof(double));
  if ((aux1d == NULL) || (aux2d == NULL))
    {
      if (p->incx != NULL) free(p->incx);
      if (aux1d != NULL) fftwf_free(aux1d);
      if (aux2d != NULL) fftwf_free(aux2d);
      free(p);
      return NULL;
    }

  srcftd(init,p->d,p->x,p->incx,p->incmx[0],
         p->y,p->incy,p->incmy[0],
         p->n,p->m[0],p->isign,p->scale,
         aux1d,&p->naux1,aux2d,&p->naux2);

  fftwf_free(aux1d);
  fftwf_free(aux2d);
  errstr(&ierno,s2015);
  /* allocate aux1 and aux2 arrays                            */
  p->aux1 = (double *) fftwf_malloc((_ESVINT64)p->naux1*sizeof(double));
  p->aux2 = (double *) fftwf_malloc((_ESVINT64)p->naux2*sizeof(double));
  if ((p->aux1 == NULL) || (p->aux2 == NULL))
    {
      if (p->incx != NULL) free(p->incx);
      if (p->aux1 != NULL) fftwf_free(p->aux1);
      if (p->aux2 != NULL) fftwf_free(p->aux2);
      free(p);
      return NULL;
    }
  /* initialization of srcftd                                  */

  srcftd(init,p->d,p->x,p->incx,p->incmx[0],
         p->y,p->incy,p->incmy[0],
         p->n,p->m[0],p->isign,p->scale,
         p->aux1,&p->naux1,p->aux2,&p->naux2);

  return p; 
}


fftwf_plan fftwf_plan_dft_r2c_2d(int nx, int ny, 
			       float *in, fftwf_complex *out, 
			       unsigned flags)
{
  fftwf_plan p;
  double *aux1d, *aux2d;
  int (*iusadr) (); 
  int ierno,inoal,inomes,itrace,irange,dummy;
  char s2015[8];
  int d;
  int init;
  init = 1;
  if ((flags & FFTW_NO_SIMD) == FFTW_NO_SIMD) init = 2;
  if ((flags & FFTW_UNALIGNED) == FFTW_UNALIGNED) init = 2;
  /* allocate the structure fftwf_esv_plan_s                  */
  p = (fftwf_plan) malloc(sizeof(struct fftwf_esv_plan_s)); 
  if (p == NULL) return NULL;
  /* fill the structure fftwf_esv_plan_s                      */
  p->howmany_rank = 1;
  p->mm = NULL;
  /* type of the transform: complex-to-complex,
     complex-to-real, or real-to-complex                      */
  p->type = ESV_R2C;
  /* dimension of the transform                               */
  p->d = 2;
  /* input array                                              */
  p->x = in;
  /* output array                                             */
  p->y = out;
  /* scale factor                                             */
  p->scale = 1.0;
  /* allocate all the integer arrays at the same time         */
  p->incx = (int *) malloc(11*sizeof(int));
  if (p->incx == NULL)
    {
      free(p);
      return NULL;
    }
  /* strides for the input array                              */
  /* strides for the output array                             */
  /* length of the transform                                  */
  /* number of arrays to be transformed                       */
  /* direction of the transform                               */
  p->incmx = &(p->incx[2]);
  p->incy = &(p->incmx[1]);
  p->incmy = &(p->incy[2]);
  p->n = &(p->incmy[1]);
  p->m = &(p->n[2]);
  p->isign = &(p->m[1]);

  /* size one dft                                             */
  d = 0;
  if (nx != 1)
    {
      d = d+1;
    }
  if (ny != 1)
    {
      d = d+1;
    }

  if (d == 0) 
    {
      p->d = 0;
      p->m[0] = 1;
      p->incmy[0] = 1;
      if ((void*) in == (void*) out)
	{
	  p->incmx[0] = 2;
	}
      else
	{
	  p->incmx[0] = 1;
	}
      return p;
    }

 /* n[i] != 1 at least for an i                */
  p->n[0] = ny;
  p->n[1] = nx;
  p->incx[0] = 1;
  p->incy[0] = 1;
  p->isign[0] = 1;
  p->isign[1] = 1;

  if ((void*) in == (void*) out)
    {
      p->incx[1] = 2*(p->n[0]/2+1);
    }
  else
    {
      p->incx[1] = p->n[0];
    }
  p->incy[1] = p->n[0]/2+1;


  p->incmx[0] = p->n[1]*p->incx[1];
  p->incmy[0] = p->n[1]*p->incy[1];
 
  p->m[0] = 1;
  /* determine size of aux1 and aux2 arrays                   */
  einfo(0,&dummy,&dummy);
  iusadr = enotrm;
  ierno = 2015;
  inoal = 0;
  inomes = -1;
  itrace = 0;
  irange = 2015;
  errsav(&ierno,s2015);
  errset(&ierno,&inoal,&inomes,&itrace,&iusadr,&irange);
  p->naux1 = 19;
  p->naux2 = 0;
  aux1d = (double *) fftwf_malloc((_ESVINT64)p->naux1*sizeof(double));
  aux2d = (double *) fftwf_malloc((_ESVINT64)2*sizeof(double));
  if ((aux1d == NULL) || (aux2d == NULL))
    {
      if (p->incx != NULL) free(p->incx);
      if (aux1d != NULL) fftwf_free(aux1d);
      if (aux2d != NULL) fftwf_free(aux2d);
      free(p);
      return NULL;
    }

  srcftd(init,p->d,p->x,p->incx,p->incmx[0],
         p->y,p->incy,p->incmy[0],
         p->n,p->m[0],p->isign,p->scale,
         aux1d,&p->naux1,aux2d,&p->naux2);

  fftwf_free(aux1d);
  fftwf_free(aux2d);
  errstr(&ierno,s2015);
  /* allocate aux1 and aux2 arrays                            */
  p->aux1 = (double *) fftwf_malloc((_ESVINT64)p->naux1*sizeof(double));
  p->aux2 = (double *) fftwf_malloc((_ESVINT64)p->naux2*sizeof(double));
  if ((p->aux1 == NULL) || (p->aux2 == NULL))
    {
      if (p->incx != NULL) free(p->incx);
      if (p->aux1 != NULL) fftwf_free(p->aux1);
      if (p->aux2 != NULL) fftwf_free(p->aux2);
      free(p);
      return NULL;
    }
  /* initialization of srcftd                                 */

  srcftd(init,p->d,p->x,p->incx,p->incmx[0],
         p->y,p->incy,p->incmy[0],
         p->n,p->m[0],p->isign,p->scale,
         p->aux1,&p->naux1,p->aux2,&p->naux2);

  return p; 
}

fftwf_plan fftwf_plan_dft_r2c_3d(int nx, int ny, int nz, 
			       float *in, fftwf_complex *out, 
			       unsigned flags)
{
  fftwf_plan p;
  double *aux1d, *aux2d;
  int (*iusadr) (); 
  int ierno,inoal,inomes,itrace,irange,dummy;
  char s2015[8];
  int d;
  int init;
  init = 1;
  if ((flags & FFTW_NO_SIMD) == FFTW_NO_SIMD) init = 2;
  if ((flags & FFTW_UNALIGNED) == FFTW_UNALIGNED) init = 2; 
  /* allocate the structure fftwf_esv_plan_s                   */
  p = (fftwf_plan) malloc(sizeof(struct fftwf_esv_plan_s)); 
  if (p == NULL) return NULL;
  /* fill the structure fftwf_esv_plan_s                       */
  p->howmany_rank = 1;
  p->mm = NULL;
  /* type of the transform: complex-to-complex,
     complex-to-real, or real-to-complex                      */
  p->type = ESV_R2C;
  /* dimension of the transform                               */
  p->d = 3;
  /* input array                                              */
  p->x = in;
  /* output array                                             */
  p->y = out;
  /* scale factor                                             */
  p->scale = 1.0;
  /* allocate all the integer arrays at the same time         */
  p->incx = (int *) malloc(15*sizeof(int));
  if (p->incx == NULL)
    {
      free(p);
      return NULL;
    }
  /* strides for the input array                              */
  /* strides for the output array                             */
  /* length of the transform                                  */
  /* number of arrays to be transformed                       */
  /* direction of the transform                               */
  p->incmx = &(p->incx[3]);
  p->incy = &(p->incmx[1]);
  p->incmy = &(p->incy[3]);
  p->n = &(p->incmy[1]);
  p->m = &(p->n[3]);
  p->isign = &(p->m[1]);

  /* size one dft                                             */
  d = 0;
  if (nx != 1)
    {
      d = d+1;
    }
  if (ny != 1)
    {
      d = d+1;
    }
  if (nz != 1)
    {
      d = d+1;
    }  

  if (d == 0) 
    {
      p->d = 0;
      p->m[0] = 1;
      p->incmy[0] = 1;
      if ((void*) in == (void*) out)
	{
	  p->incmx[0] = 2;
	}
      else
	{
	  p->incmx[0] = 1;
	}
      return p;
    }

/* n[i] != 1 at least for an i                */
  p->n[0] = nz;
  p->n[1] = ny;
  p->n[2] = nx;
  p->incx[0] = 1;
  p->incy[0] = 1;
  p->isign[0] = 1;
  p->isign[1] = 1;
  p->isign[2] = 1;

  if ((void*) in == (void*) out)
    {
      p->incx[1] = 2*(p->n[0]/2+1);
    }
  else
    {
      p->incx[1] = p->n[0];
    }
  p->incy[1] = p->n[0]/2+1;

  p->incx[2] = p->n[1]*p->incx[1];
  p->incy[2] = p->n[1]*p->incy[1];

  p->incmx[0] = p->n[2]*p->incx[2];
  p->incmy[0] = p->n[2]*p->incy[2];

  p->m[0] = 1;
  /* determine size of aux1 and aux2 arrays                   */
  einfo(0,&dummy,&dummy);
  iusadr = enotrm;
  ierno = 2015;
  inoal = 0;
  inomes = -1;
  itrace = 0;
  irange = 2015;
  errsav(&ierno,s2015);
  errset(&ierno,&inoal,&inomes,&itrace,&iusadr,&irange);
  p->naux1 = 23;
  p->naux2 = 0;
  aux1d = (double *) fftwf_malloc((_ESVINT64)p->naux1*sizeof(double));
  aux2d = (double *) fftwf_malloc((_ESVINT64)2*sizeof(double));
  if ((aux1d == NULL) || (aux2d == NULL))
    {
      if (p->incx != NULL) free(p->incx);
      if (aux1d != NULL) fftwf_free(aux1d);
      if (aux2d != NULL) fftwf_free(aux2d);
      free(p);
      return NULL;
    }

  srcftd(init,p->d,p->x,p->incx,p->incmx[0],
         p->y,p->incy,p->incmy[0],
         p->n,p->m[0],p->isign,p->scale,
         aux1d,&p->naux1,aux2d,&p->naux2);

  fftwf_free(aux1d);
  fftwf_free(aux2d);
  errstr(&ierno,s2015);
  /* allocate aux1 and aux2 arrays                            */
  p->aux1 = (double *) fftwf_malloc((_ESVINT64)p->naux1*sizeof(double));
  p->aux2 = (double *) fftwf_malloc((_ESVINT64)p->naux2*sizeof(double));
  if ((p->aux1 == NULL) || (p->aux2 == NULL))
    {
      if (p->incx != NULL) free(p->incx);
      if (p->aux1 != NULL) fftwf_free(p->aux1);
      if (p->aux2 != NULL) fftwf_free(p->aux2);
      free(p);
      return NULL;
    }
  /* initialization of srcftd                                  */

  srcftd(init,p->d,p->x,p->incx,p->incmx[0],
         p->y,p->incy,p->incmy[0],
         p->n,p->m[0],p->isign,p->scale,
         p->aux1,&p->naux1,p->aux2,&p->naux2);

  return p; 
}

fftwf_plan fftwf_plan_dft_r2c(int rank, const int *n, 
			    float *in, fftwf_complex *out, 
			    unsigned flags)
{
  fftwf_plan p;
  double *aux1d, *aux2d;
  int (*iusadr) (); 
  int ierno,inoal,inomes,itrace,irange,dummy;
  char s2015[8];
  int d, i;
  int init;
  init = 1;
  if ((flags & FFTW_NO_SIMD) == FFTW_NO_SIMD) init = 2;
  if ((flags & FFTW_UNALIGNED) == FFTW_UNALIGNED) init = 2;
  /* allocate the structure fftwf_esv_plan_s                  */
  p = (fftwf_plan) malloc(sizeof(struct fftwf_esv_plan_s)); 
  if (p == NULL) return NULL;
  /* fill the structure fftwf_esv_plan_s                      */
  p->howmany_rank = 1;
  p->mm = NULL;
  /* type of the transform: complex-to-complex,
     complex-to-real, or real-to-complex                      */
  p->type = ESV_R2C;
  /* dimension of the transform                               */
  p->d = rank;
  /* input array                                              */
  p->x = in;
  /* output array                                             */
  p->y = out;
  /* scale factor                                             */
  p->scale = 1.0;
  /* allocate all the integer arrays at the same time         */
  p->incx = (int *) malloc((_ESVINT64)(4*rank+3)*sizeof(int));  
  if (p->incx == NULL)
    {
      free(p);
      return NULL;
    }
  /* strides for the input array                              */
  /* strides for the output array                             */
  /* length of the transform                                  */
  /* number of arrays to be transformed                       */
  /* direction of the transform                               */
  p->incmx = &(p->incx[rank]);
  p->incy = &(p->incmx[1]);
  p->incmy = &(p->incy[rank]);
  p->n = &(p->incmy[1]);
  p->m = &(p->n[rank]);
  p->isign = &(p->m[1]);

  /* size one dft                                             */
  d = 0;
  for (i=0; i<p->d; i++)
    {
      if (n[i] != 1)
	{
	  d = d+1;
	}
    }

  if (d == 0) 
    {
      p->d = 0;
      p->m[0] = 1;
      p->incmy[0] = 1;
      if ((void*) in == (void*) out)
	{
	  p->incmx[0] = 2;
	}
      else
	{
	  p->incmx[0] = 1;
	}
      return p;
    }

/* n[i] != 1 at least for an i                */
  for (i=0; i<p->d; i++)
    {
      p->n[i] = n[p->d-1-i];
      p->isign[i] = 1;
    }
  p->incx[0] = 1;
  p->incy[0] = 1;

  if ((void*) in == (void*) out)
    {
      p->incx[1] = 2*(p->n[0]/2+1);
    }
  else
    {
      p->incx[1] = p->n[0];
    }
  p->incy[1] = p->n[0]/2+1;

  for (i=2; i<p->d; i++)
    {
      p->incx[i] = p->n[i-1]*p->incx[i-1];
      p->incy[i] = p->n[i-1]*p->incy[i-1];
    }

  p->incmx[0] = p->n[p->d-1]*p->incx[p->d-1];
  p->incmy[0] = p->n[p->d-1]*p->incy[p->d-1];

  p->m[0] = 1;
  /* determine size of aux1 and aux2 arrays                   */
  einfo(0,&dummy,&dummy);
  iusadr = enotrm;
  ierno = 2015;
  inoal = 0;
  inomes = -1;
  itrace = 0;
  irange = 2015;
  errsav(&ierno,s2015);
  errset(&ierno,&inoal,&inomes,&itrace,&iusadr,&irange);
  p->naux1 = 11+4*p->d;
  p->naux2 = 0;
  aux1d = (double *) fftwf_malloc((_ESVINT64)p->naux1*sizeof(double));
  aux2d = (double *) fftwf_malloc((_ESVINT64)2*sizeof(double));
  if ((aux1d == NULL) || (aux2d == NULL))
    {
      if (p->incx != NULL) free(p->incx);
      if (aux1d != NULL) fftwf_free(aux1d);
      if (aux2d != NULL) fftwf_free(aux2d);
      free(p);
      return NULL;
    }

  srcftd(init,p->d,p->x,p->incx,p->incmx[0],
         p->y,p->incy,p->incmy[0],
         p->n,p->m[0],p->isign,p->scale,
         aux1d,&p->naux1,aux2d,&p->naux2);

  fftwf_free(aux1d);
  fftwf_free(aux2d);
  errstr(&ierno,s2015);
  /* allocate aux1 and aux2 arrays                            */
  p->aux1 = (double *) fftwf_malloc((_ESVINT64)p->naux1*sizeof(double));
  p->aux2 = (double *) fftwf_malloc((_ESVINT64)p->naux2*sizeof(double));
  if ((p->aux1 == NULL) || (p->aux2 == NULL))
    {
      if (p->incx != NULL) free(p->incx);
      if (p->aux1 != NULL) fftwf_free(p->aux1);
      if (p->aux2 != NULL) fftwf_free(p->aux2);
      free(p);
      return NULL;
    }
  /* initialization of srcftd                                 */

  srcftd(init,p->d,p->x,p->incx,p->incmx[0],
         p->y,p->incy,p->incmy[0],
         p->n,p->m[0],p->isign,p->scale,
         p->aux1,&p->naux1,p->aux2,&p->naux2);

  return p; 
}

fftwf_plan fftwf_plan_many_dft_r2c(int rank, const int *n, int howmany,
				 float *in, const int *inembed,
				 int istride, int idist,
				 fftwf_complex *out, const int *onembed,
				 int ostride, int odist,  
				 unsigned flags)
{
  fftwf_plan p;
  double *aux1d, *aux2d;
  int (*iusadr) (); 
  int ierno,inoal,inomes,itrace,irange,dummy;
  char s2015[8];
  int d, j, i;
  int init;
  init = 1;
  if ((flags & FFTW_NO_SIMD) == FFTW_NO_SIMD) init = 2;
  if ((flags & FFTW_UNALIGNED) == FFTW_UNALIGNED) init = 2;
  /* allocate the structure fftwf_esv_plan_s                  */
  p = (fftwf_plan) malloc(sizeof(struct fftwf_esv_plan_s)); 
  if (p == NULL) return NULL;
  /* fill the structure fftwf_esv_plan_s                      */
  p->howmany_rank = 1;
  p->mm = NULL;
  /* type of the transform: complex-to-complex,
     complex-to-real, or real-to-complex                      */
  p->type = ESV_R2C;
  /* dimension of the transform                               */
  p->d = rank;
  /* input array                                              */
  p->x = in;
  /* output array                                             */
  p->y = out;
  /* scale factor                                             */
  p->scale = 1.0;
  /* allocate all the integer arrays at the same time         */
  p->incx = (int *) malloc((_ESVINT64)(4*rank+3)*sizeof(int));
  if (p->incx == NULL)
    {
      free(p);
      return NULL;
    }

  p->incmx = &(p->incx[rank]);
  p->incy = &(p->incmx[1]);
  p->incmy = &(p->incy[rank]);
  p->n = &(p->incmy[1]);
  p->m = &(p->n[rank]);
  p->isign = &(p->m[1]);

  /* size one dft                                             */
  d = 0;
  for (i=0; i<p->d; i++)
    {
      if (n[i] != 1)
	{d = d+1;}
    }

  if (d == 0) 
    {
      p->d = 0;
      p->m[0] = howmany;
      p->incmx[0] = idist;
      p->incmy[0] = odist;
      return p;
    }

  /* n[i] != 1 at least for an i                */
  for (i=0; i<p->d; i++)
    {
      p->n[i] = n[p->d-1-i];
      p->isign[i] = 1;
    }

  /* strides for the input array                              */
  if (inembed == NULL)
    {
      p->incx[0] = istride;
      if ((void*) in == (void*) out)
	{
	  p->incx[1] = 2*istride*(p->n[0]/2+1);
	}
      else
	{
	  p->incx[1] = istride*p->n[0];
	}
      for (i=2; i<p->d; i++)
	{
	  p->incx[i] = p->n[i-1]*p->incx[i-1];
	}
    }
  else
    {
      p->incx[0] = istride;
      for (i=1; i<rank; i++)
	{
	  p->incx[i] = inembed[rank-i]*p->incx[i-1];
	}
    }
  p->incmx[0] = idist;
  /* strides for the output array                             */
  if (onembed == NULL)
    {
      p->incy[0] = ostride;
      p->incy[1] = ostride*(p->n[0]/2+1);
      for (i=2; i<p->d; i++)
	{
	  p->incy[i] = p->n[i-1]*p->incy[i-1];
	}
    }
  else
    {
      p->incy[0] = ostride;
      for (i=1; i<p->d; i++)
	{
	  p->incy[i] = onembed[rank-i]*p->incy[i-1];
	}
    }
  p->incmy[0] = odist;
  /* number of arrays to be transformed                       */
  p->m[0] = howmany;
  /* determine size of aux1 and aux2 arrays                   */
  einfo(0,&dummy,&dummy);
  iusadr = enotrm;
  ierno = 2015;
  inoal = 0;
  inomes = -1;
  itrace = 0;
  irange = 2015;
  errsav(&ierno,s2015);
  errset(&ierno,&inoal,&inomes,&itrace,&iusadr,&irange);
  p->naux1 = 11+4*p->d;
  p->naux2 = 0;
  aux1d = (double *) fftwf_malloc((_ESVINT64)p->naux1*sizeof(double));
  aux2d = (double *) fftwf_malloc((_ESVINT64)2*sizeof(double));
  if ((aux1d == NULL) || (aux2d == NULL))
    {
      if (p->incx != NULL) free(p->incx);
      if (aux1d != NULL) fftwf_free(aux1d);
      if (aux2d != NULL) fftwf_free(aux2d);
      free(p);
      return NULL;
    }

  srcftd(init,p->d,p->x,p->incx,p->incmx[0],
         p->y,p->incy,p->incmy[0],
         p->n,p->m[0],p->isign,p->scale,
         aux1d,&p->naux1,aux2d,&p->naux2);

  fftwf_free(aux1d);
  fftwf_free(aux2d);
  errstr(&ierno,s2015);
  /* allocate aux1 and aux2 arrays                            */
  p->aux1 = (double *) fftwf_malloc((_ESVINT64)p->naux1*sizeof(double));
  p->aux2 = (double *) fftwf_malloc((_ESVINT64)p->naux2*sizeof(double));
  if ((p->aux1 == NULL) || (p->aux2 == NULL))
    {
      if (p->incx != NULL) free(p->incx);
      if (p->aux1 != NULL) fftwf_free(p->aux1);
      if (p->aux2 != NULL) fftwf_free(p->aux2);
      free(p);
      return NULL;
    }
  /* initialization of srcftd                                 */

  srcftd(init,p->d,p->x,p->incx,p->incmx[0],
         p->y,p->incy,p->incmy[0],
         p->n,p->m[0],p->isign,p->scale,
         p->aux1,&p->naux1,p->aux2,&p->naux2);

  return p; 
}

fftwf_plan fftwf_plan_guru_dft_r2c(int rank, const fftwf_iodim *dims,
				 int howmany_rank, 
				 const fftwf_iodim *howmany_dims,
				 float *in, fftwf_complex *out,
				 unsigned flags)
{
  fftwf_plan p;
  double *aux1d, *aux2d;
  int (*iusadr) (); 
  int ierno,inoal,inomes,itrace,irange,dummy;
  char s2015[8];
  int size;
  int d, i, stride;
  int init;
  init = 1;
  if ((flags & FFTW_NO_SIMD) == FFTW_NO_SIMD) init = 2;
  if ((flags & FFTW_UNALIGNED) == FFTW_UNALIGNED) init = 2;
  /* allocate the structure fftwf_esv_plan_s                  */
  p = (fftwf_plan) malloc(sizeof(struct fftwf_esv_plan_s)); 
  if (p == NULL) return NULL;
  /* fill the structure fftwf_esv_plan_s                      */
  /* type of the transform: complex-to-complex,
     complex-to-real, or real-to-complex                      */
  p->type = ESV_R2C;
  /* dimension of the transform                               */
  p->d = rank;
  /* input array                                              */
  p->x = in;
  /* output array                                             */
  p->y = out;
  /* scale factor                                             */
  p->scale = 1.0;
  /* allocate all the integer arrays at the same time         */
  if (howmany_rank > 1)
    {
      size = 4*(rank + howmany_rank);
    }
  else
    {
      size = 4*rank + 3;
    }
  p->incx = (int *) malloc((_ESVINT64)size*sizeof(int));
  if (p->incx == NULL)
    {
      free(p);
      return NULL;
    }
  /* strides for the input array                              */
  /* strides for the output array                             */
  /* length of the transform                                  */
  /* direction of the transform                               */
  p->incy = &(p->incx[rank]);
  p->n = &(p->incy[rank]);
  p->isign = &(p->n[rank]);
  for (i=0; i<rank; i++)
    {
      p->incx[i] = dims[rank-1-i].is;
      p->incy[i] = dims[rank-1-i].os;
      p->n[i] = dims[rank-1-i].n;
      p->isign[i] = 1;
    }
  /* p->incmx, p->incmy, p->m, p->howmany_rank, and p->mm
     depend on the parameter howmany_rank                     */
  if (howmany_rank == 0)
    {
      p->howmany_rank = 1;
      p->mm = NULL;
      p->m = &(p->isign[rank]);
      p->incmx = &(p->m[1]);
      p->incmy = &(p->incmx[1]);

      if (rank == 0)
	{
	  p->m[0] = 1;
	  p->incmy[0] = 1;
	  if ((void*) in == (void*) out)
	    {
	      p->incmx[0] = 2;
	    }
	  else
	    {
	      p->incmx[0] = 1;
	    }
	}
      else
	{
	  p->m[0] = 1;
	  if (rank > 1)
	    {
	      p->incmx[0] = p->n[rank-1]*p->incx[rank-1];
	      p->incmy[0] = p->n[rank-1]*p->incy[rank-1];
	    }
	  else
	    {
	      p->incmy[0] = (p->n[0]/2+1)*p->incy[0];	 
	      if ((void*) in == (void*) out)
		{
		  p->incmx[0] = 2*(p->n[0]/2+1)*p->incx[0];	
		}
	      else
		{
		  p->incmx[0] = p->n[0]*p->incx[0];
		}
	    }
	}
    }
  else if (howmany_rank == 1)
    {
      p->howmany_rank = 1;
      p->mm = NULL;
      p->m = &(p->isign[rank]);
      p->incmx = &(p->m[1]);
      p->incmy = &(p->incmx[1]);
 
      p->m[0] = howmany_dims[0].n;
      p->incmx[0] = howmany_dims[0].is;
      p->incmy[0] = howmany_dims[0].os;
    }
  else
    {
      p->howmany_rank = howmany_rank;
      p->m = &(p->isign[rank]);
      p->incmx = &(p->m[howmany_rank]);
      p->incmy = &(p->incmx[howmany_rank]);
      p->mm = &(p->incmy[howmany_rank]); 

      p->m[0] = howmany_dims[0].n;
      p->incmx[0] = howmany_dims[0].is;
      p->incmy[0] = howmany_dims[0].os;
      p->mm[0] = 1;
      for (i=1; i<howmany_rank; i++)
	{
	  p->m[i] = howmany_dims[i].n;
	  p->incmx[i] = howmany_dims[i].is;
	  p->incmy[i] = howmany_dims[i].os;

	  p->mm[i] = p->mm[0];
	  p->mm[0] = p->mm[0]*p->m[i];
	}
      /* here also we have to decide whether we will use      */
      /* SIMD performance enhancements                        */
      for (i=1; i<howmany_rank; i++)
	{
	  stride = p->incmx[i]/4;
	  if (p->incmx[i] != 4*stride) init = 2;
	  stride = p->incmy[i]/2;
	  if (p->incmy[i] != 2*stride) init = 2;
	}
    }


  /* size one dft                                             */
  d = 0;
  for (i=0; i<p->d; i++)
    {
      if (p->n[i] != 1)
	{d = d+1;}
    }

  if (d == 0) 
    {
      p->d = 0;
      return p;
    }

  /* determine size of aux1 and aux2 arrays                   */
  einfo(0,&dummy,&dummy);
  iusadr = enotrm;
  ierno = 2015;
  inoal = 0;
  inomes = -1;
  itrace = 0;
  irange = 2015;
  errsav(&ierno,s2015);
  errset(&ierno,&inoal,&inomes,&itrace,&iusadr,&irange);
  p->naux1 = 11+4*p->d;
  p->naux2 = 0;
  aux1d = (double *) fftwf_malloc((_ESVINT64)p->naux1*sizeof(double));
  aux2d = (double *) fftwf_malloc((_ESVINT64)2*sizeof(double));
  if ((aux1d == NULL) || (aux2d == NULL))
    {
      if (p->incx != NULL) free(p->incx);
      if (aux1d != NULL) fftwf_free(aux1d);
      if (aux2d != NULL) fftwf_free(aux2d);
      free(p);
      return NULL;
    }

  srcftd(init,p->d,p->x,p->incx,p->incmx[0],
         p->y,p->incy,p->incmy[0],
         p->n,p->m[0],p->isign,p->scale,
         aux1d,&p->naux1,aux2d,&p->naux2);

  fftwf_free(aux1d);
  fftwf_free(aux2d);
  errstr(&ierno,s2015);
  /* allocate aux1 and aux2 arrays                            */
  p->aux1 = (double *) fftwf_malloc((_ESVINT64)p->naux1*sizeof(double));
  p->aux2 = (double *) fftwf_malloc((_ESVINT64)p->naux2*sizeof(double));
  if ((p->aux1 == NULL) || (p->aux2 == NULL))
    {
      if (p->incx != NULL) free(p->incx);
      if (p->aux1 != NULL) fftwf_free(p->aux1);
      if (p->aux2 != NULL) fftwf_free(p->aux2);
      free(p);
      return NULL;
    }
  /* initialization of srcftd                                 */

  srcftd(init,p->d,p->x,p->incx,p->incmx[0],
         p->y,p->incy,p->incmy[0],
         p->n,p->m[0],p->isign,p->scale,
         p->aux1,&p->naux1,p->aux2,&p->naux2);

  return p; 
}

/* C2R ----------------------------------------------------------- */

fftwf_plan fftwf_plan_dft_c2r_1d(int n, fftwf_complex *in, 
			       float *out, unsigned flags)
{
  fftwf_plan p;
  double *aux1d, *aux2d;
  int (*iusadr) (); 
  int ierno,inoal,inomes,itrace,irange,dummy;
  char s2015[8];
  int init;
  init = 1;
  if ((flags & FFTW_NO_SIMD) == FFTW_NO_SIMD) init = 2;
  if ((flags & FFTW_UNALIGNED) == FFTW_UNALIGNED) init = 2;
  /* allocate the structure fftwf_esv_plan_s                  */
  p = (fftwf_plan) malloc(sizeof(struct fftwf_esv_plan_s)); 
  if (p == NULL) return NULL;
  /* fill the structure fftwf_esv_plan_s                      */
  p->howmany_rank = 1;
  p->mm = NULL;
  /* type of the transform: complex-to-complex,
     complex-to-real, or real-to-complex                      */
  p->type = ESV_C2R;
  /* dimension of the transform                               */
  p->d = 1;
  /* input array                                              */
  p->x = in;
  /* output array                                             */
  p->y = out;
  /* scale factor                                             */
  p->scale = 1.0;
  /* allocate all the integer arrays at the same time         */
  p->incx = (int *) malloc(7*sizeof(int));
  if (p->incx == NULL)
    {
      free(p);
      return NULL;
    }
  /* strides for the input array                              */
  p->incx[0] = 1;
  p->incmx = &(p->incx[1]);
  p->incmx[0] = n/2+1;
  /* strides for the output array                             */
  p->incy = &(p->incmx[1]);
  p->incy[0] = 1;
  p->incmy = &(p->incy[1]);
  if ((void*) in == (void*) out)
    {
      p->incmy[0] = 2*(n/2+1);
    }
  else
    {
      p->incmy[0] = n;
    }
  /* length of the transform                                  */
  p->n = &(p->incmy[1]);
  p->n[0] = n;
  /* number of arrays to be transformed                       */
  p->m = &(p->n[1]);
  p->m[0] = 1;
  /* direction of the transform                               */
  p->isign = &(p->m[1]);
  p->isign[0] = -1;
  /* size one dft                                             */
  if (n == 1)
    {
      p->d = 0;
      return p;
    }
  /* determine size of aux1 and aux2 arrays                   */
  einfo(0,&dummy,&dummy);
  iusadr = enotrm;
  ierno = 2015;
  inoal = 0;
  inomes = -1;
  itrace = 0;
  irange = 2015;
  errsav(&ierno,s2015);
  errset(&ierno,&inoal,&inomes,&itrace,&iusadr,&irange);
  p->naux1 = 15;
  p->naux2 = 0;
  aux1d = (double *) fftwf_malloc((_ESVINT64)p->naux1*sizeof(double));
  aux2d = (double *) fftwf_malloc((_ESVINT64)2*sizeof(double));
  if ((aux1d == NULL) || (aux2d == NULL))
    {
      if (p->incx != NULL) free(p->incx);
      if (aux1d != NULL) fftwf_free(aux1d);
      if (aux2d != NULL) fftwf_free(aux2d);
      free(p);
      return NULL;
    }
  fftwf_dft_c2r_x(init,p->d,p->x,p->incx,p->incmx[0],
		  p->y,p->incy,p->incmy[0],
		  p->n,p->m[0],p->isign,p->scale,
		  aux1d,&p->naux1,aux2d,&p->naux2);
  fftwf_free(aux1d);
  fftwf_free(aux2d);
  errstr(&ierno,s2015);
  /* allocate aux1 and aux2 arrays                            */
  p->aux1 = (double *) fftwf_malloc((_ESVINT64)p->naux1*sizeof(double));
  p->aux2 = (double *) fftwf_malloc((_ESVINT64)p->naux2*sizeof(double));
  if ((p->aux1 == NULL) || (p->aux2 == NULL))
    {
      if (p->incx != NULL) free(p->incx);
      if (p->aux1 != NULL) fftwf_free(p->aux1);
      if (p->aux2 != NULL) fftwf_free(p->aux2);
      free(p);
      return NULL;
    }
  /* initialization of fftwf_dft_c2r_x                        */
  fftwf_dft_c2r_x(init,p->d,p->x,p->incx,p->incmx[0],
		  p->y,p->incy,p->incmy[0],
		  p->n,p->m[0],p->isign,p->scale,
		  p->aux1,&p->naux1,p->aux2,&p->naux2);
  return p;  
  return p; 
}

fftwf_plan fftwf_plan_dft_c2r_2d(int nx, int ny, 
			       fftwf_complex *in, float *out, 
			       unsigned flags)
{
  fftwf_plan p;
  double *aux1d, *aux2d; 
  int (*iusadr) (); 
  int ierno,inoal,inomes,itrace,irange,dummy;
  char s2015[8];
  int d;
  int init;
  init = 1;
  if ((flags & FFTW_NO_SIMD) == FFTW_NO_SIMD) init = 2;
  if ((flags & FFTW_UNALIGNED) == FFTW_UNALIGNED) init = 2;
  if ((flags & FFTW_PRESERVE_INPUT) == FFTW_PRESERVE_INPUT) return NULL;
  /* allocate the structure fftwf_esv_plan_s                  */
  p = (fftwf_plan) malloc(sizeof(struct fftwf_esv_plan_s)); 
  if (p == NULL) return NULL;
  /* fill the structure fftwf_esv_plan_s                      */
  p->howmany_rank = 1;
  p->mm = NULL;
  /* type of the transform: complex-to-complex,
     complex-to-real, or real-to-complex                      */
  p->type = ESV_C2R;
  /* dimension of the transform                               */
  p->d = 2;
  /* input array                                              */
  p->x = in;
  /* output array                                             */
  p->y = out;
  /* scale factor                                             */
  p->scale = 1.0;
  /* allocate all the integer arrays at the same time         */
  p->incx = (int *) malloc(11*sizeof(int));
  if (p->incx == NULL)
    {
      free(p);
      return NULL;
    }
  /* strides for the input array                              */
  /* strides for the output array                             */
  /* length of the transform                                  */
  /* number of arrays to be transformed                       */
  /* direction of the transform                               */
  p->incmx = &(p->incx[2]);
  p->incy = &(p->incmx[1]);
  p->incmy = &(p->incy[2]);
  p->n = &(p->incmy[1]);
  p->m = &(p->n[2]);
  p->isign = &(p->m[1]);

  /* size one dft                                             */
  d = 0;
  if (nx != 1)
    {d = d+1;}
  if (ny != 1)
    {d = d+1;}

  if (d == 0) 
    {
      p->d = 0;
      p->m[0] = 1;
      p->incmx[0] = 1;
      if ((void*) in == (void*) out)
	{
	  p->incmy[0] = 2;
	}
      else
	{
	  p->incmy[0] = 1;
	}
      return p;
    }
  /*  n[i] != 1 at least for an i                 */
  p->n[0] = ny;
  p->n[1] = nx;
  p->incx[0] = 1;
  p->incy[0] = 1;
  p->isign[0] = -1;
  p->isign[1] = -1;

  p->incx[1] = p->n[0]/2+1;
  if ((void*) in == (void*) out)
    {
      p->incy[1] = 2*(p->n[0]/2+1);
    } 
  else
    {
      p->incy[1] = p->n[0];
    }
  
  p->incmx[0] = p->n[1]*p->incx[1];
  p->incmy[0] = p->n[1]*p->incy[1];

  p->m[0] = 1;
  /* determine size of aux1 and aux2 arrays                   */
  einfo(0,&dummy,&dummy);
  iusadr = enotrm;
  ierno = 2015;
  inoal = 0;
  inomes = -1;
  itrace = 0;
  irange = 2015;
  errsav(&ierno,s2015);
  errset(&ierno,&inoal,&inomes,&itrace,&iusadr,&irange);
  p->naux1 = 15;
  p->naux2 = 0;
  aux1d = (double *) fftwf_malloc((_ESVINT64)p->naux1*sizeof(double));
  aux2d = (double *) fftwf_malloc((_ESVINT64)2*sizeof(double));
  if ((aux1d == NULL) || (aux2d == NULL))
    {
      if (p->incx != NULL) free(p->incx);
      if (aux1d != NULL) fftwf_free(aux1d);
      if (aux2d != NULL) fftwf_free(aux2d);
      free(p);
      return NULL;
    }
  fftwf_dft_c2r_x(init,p->d,p->x,p->incx,p->incmx[0],
		  p->y,p->incy,p->incmy[0],
		  p->n,p->m[0],p->isign,p->scale,
		  aux1d,&p->naux1,aux2d,&p->naux2);
  fftwf_free(aux1d);
  fftwf_free(aux2d);
  errstr(&ierno,s2015);
  /* allocate aux1 and aux2 arrays                            */
  p->aux1 = (double *) fftwf_malloc((_ESVINT64)p->naux1*sizeof(double));
  p->aux2 = (double *) fftwf_malloc((_ESVINT64)p->naux2*sizeof(double));
  if ((p->aux1 == NULL) || (p->aux2 == NULL))
    {
      if (p->incx != NULL) free(p->incx);
      if (p->aux1 != NULL) fftwf_free(p->aux1);
      if (p->aux2 != NULL) fftwf_free(p->aux2);
      free(p);
      return NULL;
    }
  /* initialization of fftwf_dft_c2r_x                       */
  fftwf_dft_c2r_x(init,p->d,p->x,p->incx,p->incmx[0],
		  p->y,p->incy,p->incmy[0],
		  p->n,p->m[0],p->isign,p->scale,
		  p->aux1,&p->naux1,p->aux2,&p->naux2);
  return p; 
}

fftwf_plan fftwf_plan_dft_c2r_3d(int nx, int ny, int nz, 
			       fftwf_complex *in, float *out, 
			       unsigned flags)
{
  fftwf_plan p;
  double *aux1d, *aux2d;
  int (*iusadr) (); 
  int ierno,inoal,inomes,itrace,irange,dummy;
  char s2015[8];
  int d;
  int init;
  init = 1;
  if ((flags & FFTW_NO_SIMD) == FFTW_NO_SIMD) init = 2;
  if ((flags & FFTW_UNALIGNED) == FFTW_UNALIGNED) init = 2;
  if ((flags & FFTW_PRESERVE_INPUT) == FFTW_PRESERVE_INPUT) return NULL;
  /* allocate the structure fftwf_esv_plan_s                   */
  p = (fftwf_plan) malloc(sizeof(struct fftwf_esv_plan_s)); 
  if (p == NULL) return NULL;
  /* fill the structure fftwf_esv_plan_s                       */
  p->howmany_rank = 1;
  p->mm = NULL;
  /* type of the transform: complex-to-complex,
     complex-to-real, or real-to-complex                      */
  p->type = ESV_C2R;
  /* dimension of the transform                               */
  p->d = 3;
  /* input array                                              */
  p->x = in;
  /* output array                                             */
  p->y = out;
  /* scale factor                                             */
  p->scale = 1.0;
  /* allocate all the integer arrays at the same time         */
  p->incx = (int *) malloc(15*sizeof(int));
  if (p->incx == NULL)
    {
      free(p);
      return NULL;
    }
  /* strides for the input array                              */
  /* strides for the output array                             */
  /* length of the transform                                  */
  /* number of arrays to be transformed                       */
  /* direction of the transform                               */
  p->incmx = &(p->incx[3]);
  p->incy = &(p->incmx[1]);
  p->incmy = &(p->incy[3]);
  p->n = &(p->incmy[1]);
  p->m = &(p->n[3]);
  p->isign = &(p->m[1]);

  /* size one dft                                             */
  d = 0;
  if (nx != 1)
    {
      d = d+1;
    }
  if (ny != 1)
    {
      d = d+1;
    }
  if (nz != 1)
    {
      d = d+1;
    }

  if (d == 0) 
    {
      p->d = 0;
      p->m[0] = 1;
      p->incmx[0] = 1;
      if ((void*) in == (void*) out)
	{
	  p->incmy[0] = 2;
	}
      else
	{
	  p->incmy[0] = 1;
	}
      return p;
    }

  /*  n[i] != 1 at least for an i                 */
  p->n[0] = nz;
  p->n[1] = ny;
  p->n[2] = nx;
  p->incx[0] = 1;
  p->incy[0] = 1;
  p->isign[0] = -1;
  p->isign[1] = -1;
  p->isign[2] = -1;

  p->incx[1] = p->n[0]/2+1;
  if ((void*) in == (void*) out)
    {
      p->incy[1] = 2*(p->n[0]/2+1);
    } 
  else
    {
      p->incy[1] = p->n[0];
    }
  
  p->incx[2] = p->n[1]*p->incx[1];
  p->incy[2] = p->n[1]*p->incy[1];

  p->incmx[0] = p->n[2]*p->incx[2];
  p->incmy[0] = p->n[2]*p->incy[2];

  p->m[0] = 1;
  /* determine size of aux1 and aux2 arrays                   */
  einfo(0,&dummy,&dummy);
  iusadr = enotrm;
  ierno = 2015;
  inoal = 0;
  inomes = -1;
  itrace = 0;
  irange = 2015;
  errsav(&ierno,s2015);
  errset(&ierno,&inoal,&inomes,&itrace,&iusadr,&irange);
  p->naux1 = 22;
  p->naux2 = 0;
  aux1d = (double *) fftwf_malloc((_ESVINT64)p->naux1*sizeof(double));
  aux2d = (double *) fftwf_malloc((_ESVINT64)2*sizeof(double));
  if ((aux1d == NULL) || (aux2d == NULL))
    {
      if (p->incx != NULL) free(p->incx);
      if (aux1d != NULL) fftwf_free(aux1d);
      if (aux2d != NULL) fftwf_free(aux2d);
      free(p);
      return NULL;
    }
  fftwf_dft_c2r_x(init,p->d,p->x,p->incx,p->incmx[0],
		  p->y,p->incy,p->incmy[0],
		  p->n,p->m[0],p->isign,p->scale,
		  aux1d,&p->naux1,aux2d,&p->naux2);
  fftwf_free(aux1d);
  fftwf_free(aux2d);
  errstr(&ierno,s2015);
  /* allocate aux1 and aux2 arrays                            */
  p->aux1 = (double *) fftwf_malloc((_ESVINT64)p->naux1*sizeof(double));
  p->aux2 = (double *) fftwf_malloc((_ESVINT64)p->naux2*sizeof(double));
  if ((p->aux1 == NULL) || (p->aux2 == NULL))
    {
      if (p->incx != NULL) free(p->incx);
      if (p->aux1 != NULL) fftwf_free(p->aux1);
      if (p->aux2 != NULL) fftwf_free(p->aux2);
      free(p);
      return NULL;
    }
  /* initialization of fftwf_dft_c2r_x                      */
  fftwf_dft_c2r_x(init,p->d,p->x,p->incx,p->incmx[0],
		 p->y,p->incy,p->incmy[0],
		 p->n,p->m[0],p->isign,p->scale,
		 p->aux1,&p->naux1,p->aux2,&p->naux2);
  return p; 
}

fftwf_plan fftwf_plan_dft_c2r(int rank, const int *n, 
			    fftwf_complex *in, float *out, 
			    unsigned flags)
{
  fftwf_plan p;
  double *aux1d, *aux2d;
  int (*iusadr) (); 
  int ierno,inoal,inomes,itrace,irange,dummy;
  char s2015[8];
  int d, i;
  int init;
  init = 1;
  if ((flags & FFTW_NO_SIMD) == FFTW_NO_SIMD) init = 2;
  if ((flags & FFTW_UNALIGNED) == FFTW_UNALIGNED) init = 2;
  if (rank > 1)
    {if ((flags & FFTW_PRESERVE_INPUT) == FFTW_PRESERVE_INPUT) return NULL;}
  /* allocate the structure fftwf_esv_plan_s                  */
  p = (fftwf_plan) malloc(sizeof(struct fftwf_esv_plan_s)); 
  if (p == NULL) return NULL;
  /* fill the structure fftwf_esv_plan_s                      */
  p->howmany_rank = 1;
  p->mm = NULL;
  /* type of the transform: complex-to-complex,
     complex-to-real, or real-to-complex                      */
  p->type = ESV_C2R;
  /* dimension of the transform                               */
  p->d = rank;
  /* input array                                              */
  p->x = in;
  /* output array                                             */
  p->y = out;
  /* scale factor                                             */
  p->scale = 1.0;
  /* allocate all the integer arrays at the same time         */
  p->incx = (int *) malloc((_ESVINT64)(4*rank+3)*sizeof(int));  
  if (p->incx == NULL)
    {
      free(p);
      return NULL;
    }

  /* strides for the input array                              */
  /* strides for the output array                             */
  /* length of the transform                                  */
  /* number of arrays to be transformed                       */
  /* direction of the transform                               */
  p->incmx = &(p->incx[rank]);
  p->incy = &(p->incmx[1]);
  p->incmy = &(p->incy[rank]);
  p->n = &(p->incmy[1]);
  p->m = &(p->n[rank]);
  p->isign = &(p->m[1]);

  /* size one dft                                             */
  d = 0;
  for (i=0; i<p->d; i++)
    {
      if (n[i] != 1)
	{d = d+1;}
    }

  if (d == 0) 
    {
      p->d = 0;
      p->m[0] = 1;
      p->incmx[0] = 1;
      if ((void*) in == (void*) out)
	{
	  p->incmy[0] = 2;
	}
      else
	{
	  p->incmy[0] = 1;
	}
      return p;
    }

/* n[i] != 1 at least for an i                */
  for (i=0; i<p->d; i++)
    {
      p->n[i] = n[p->d-1-i];
      p->isign[i] = -1;
    }
  p->incx[0] = 1;
  p->incy[0] = 1;

  p->incx[1] = p->n[0]/2+1;
  if ((void*) in == (void*) out)
    {
      p->incy[1] = 2*(p->n[0]/2+1);
    } 
  else
    {
      p->incy[1] = p->n[0];
    }
  
  for (i=2; i<p->d; i++)
    {
      p->incx[i] = p->n[i-1]*p->incx[i-1];
      p->incy[i] = p->n[i-1]*p->incy[i-1];
    }

  p->incmx[0] = p->n[p->d-1]*p->incx[p->d-1];
  p->incmy[0] = p->n[p->d-1]*p->incy[p->d-1];

  p->m[0] = 1;
  /* determine size of aux1 and aux2 arrays                   */
  einfo(0,&dummy,&dummy);
  iusadr = enotrm;
  ierno = 2015;
  inoal = 0;
  inomes = -1;
  itrace = 0;
  irange = 2015;
  errsav(&ierno,s2015);
  errset(&ierno,&inoal,&inomes,&itrace,&iusadr,&irange);
  p->naux1 = 1+7*p->d;
  if (p->naux1 < 15) p->naux1=15;
  p->naux2 = 0;
  aux1d = (double *) fftwf_malloc((_ESVINT64)p->naux1*sizeof(double));
  aux2d = (double *) fftwf_malloc((_ESVINT64)2*sizeof(double));
  if ((aux1d == NULL) || (aux2d == NULL))
    {
      if (p->incx != NULL) free(p->incx);
      if (aux1d != NULL) fftwf_free(aux1d);
      if (aux2d != NULL) fftwf_free(aux2d);
      free(p);
      return NULL;
    }
  fftwf_dft_c2r_x(init,p->d,p->x,p->incx,p->incmx[0],
		  p->y,p->incy,p->incmy[0],
		  p->n,p->m[0],p->isign,p->scale,
		  aux1d,&p->naux1,aux2d,&p->naux2);
  fftwf_free(aux1d);
  fftwf_free(aux2d);
  errstr(&ierno,s2015);
  /* allocate aux1 and aux2 arrays                            */
  p->aux1 = (double *) fftwf_malloc((_ESVINT64)p->naux1*sizeof(double));
  p->aux2 = (double *) fftwf_malloc((_ESVINT64)p->naux2*sizeof(double));
  if ((p->aux1 == NULL) || (p->aux2 == NULL))
    {
      if (p->incx != NULL) free(p->incx);
      if (p->aux1 != NULL) fftwf_free(p->aux1);
      if (p->aux2 != NULL) fftwf_free(p->aux2);
      free(p);
      return NULL;
    }
  /* initialization of fftwf_dft_c2r_x                         */
  fftwf_dft_c2r_x(init,p->d,p->x,p->incx,p->incmx[0],
		  p->y,p->incy,p->incmy[0],
		  p->n,p->m[0],p->isign,p->scale,
		  p->aux1,&p->naux1,p->aux2,&p->naux2);
  return p; 
}

fftwf_plan fftwf_plan_many_dft_c2r(int rank, const int *n, int howmany,
				 fftwf_complex *in, const int *inembed,
				 int istride, int idist,
				 float *out, const int *onembed,
				 int ostride, int odist,  
				 int unsigned flags)
{
  fftwf_plan p;
  double *aux1d, *aux2d;
  int (*iusadr) (); 
  int ierno,inoal,inomes,itrace,irange,dummy;
  char s2015[8];
  int d, j, i;
  int init;
  init = 1;
  if ((flags & FFTW_NO_SIMD) == FFTW_NO_SIMD) init = 2;
  if ((flags & FFTW_UNALIGNED) == FFTW_UNALIGNED) init = 2;
  if (rank > 1)
    {if ((flags & FFTW_PRESERVE_INPUT) == FFTW_PRESERVE_INPUT) return NULL;}
  /* allocate the structure fftwf_esv_plan_s                  */
  p = (fftwf_plan) malloc(sizeof(struct fftwf_esv_plan_s)); 
  if (p == NULL) return NULL;
  /* fill the structure fftwf_esv_plan_s                      */
  p->howmany_rank = 1;
  p->mm = NULL;
  /* type of the transform: complex-to-complex,
     complex-to-real, or real-to-complex                      */
  p->type = ESV_C2R;
  /* dimension of the transform                               */
  p->d = rank;
  /* input array                                              */
  p->x = in;
  /* output array                                             */
  p->y = out;
  /* scale factor                                             */
  p->scale = 1.0;
  /* allocate all the integer arrays at the same time         */
  p->incx = (int *) malloc((_ESVINT64)(4*rank+3)*sizeof(int));
  if (p->incx == NULL)
    {
      free(p);
      return NULL;
    }

  p->incmx = &(p->incx[rank]);
  p->incy = &(p->incmx[1]);
  p->incmy = &(p->incy[rank]);
  p->n = &(p->incmy[1]);
  p->m = &(p->n[rank]);
  p->isign = &(p->m[1]);

  /* size one dft                                             */
  d = 0;
  for (i=0; i<p->d; i++)
    {
      if (n[i] != 1)
	{d = d+1;}
    }

  if (d == 0) 
    {
      p->d = 0;
      p->m[0] = howmany;
      p->incmx[0] = idist;
      p->incmy[0] = odist;
      return p;
    }

  /* n[i] != 1 at least for an i                */
  for (i=0; i<p->d; i++)
    {
      p->n[i] = n[p->d-1-i];
      p->isign[i] = -1;
    }

  /* strides for the input array                              */
  if (inembed == NULL)
    {
      p->incx[0] = istride;
      p->incx[1] = (p->n[0]/2+1)*istride;
      for (i=2; i<p->d; i++)
	{
	  p->incx[i] = p->n[i-1]*p->incx[i-1];
	}
    }
  else
    {
      p->incx[0] = istride;
      for (i=1; i<rank; i++)
	{
	  p->incx[i] = inembed[rank-i]*p->incx[i-1];
	}
    }
  p->incmx[0] = idist;
  /* strides for the output array                             */
  if (onembed == NULL)
    {
      p->incy[0] = ostride;
      if ((void*) in == (void*) out)
	{
	  p->incy[1] = 2*(p->n[0]/2+1)*ostride;
	}
      else
	{
	  p->incy[1] = p->n[0]*ostride;
	}
      for (i=2; i<p->d; i++)
	{
	  p->incy[i] = p->n[i-1]*p->incy[i-1];
	}
    }
  else
    {
      p->incy[0] = ostride;
      for (i=1; i<rank; i++)
	{
	  p->incy[i] = onembed[rank-i]*p->incy[i-1];
	}
    }
  p->incmy[0] = odist;
  /* number of arrays to be transformed                       */
  p->m[0] = howmany;
  /* determine size of aux1 and aux2 arrays                   */
  einfo(0,&dummy,&dummy);
  iusadr = enotrm;
  ierno = 2015;
  inoal = 0;
  inomes = -1;
  itrace = 0;
  irange = 2015;
  errsav(&ierno,s2015);
  errset(&ierno,&inoal,&inomes,&itrace,&iusadr,&irange);
  p->naux1 = 1+7*p->d;
  if (p->naux1 < 15) p->naux1=15;
  p->naux2 = 0;
  aux1d = (double *) fftwf_malloc((_ESVINT64)p->naux1*sizeof(double));
  aux2d = (double *) fftwf_malloc((_ESVINT64)2*sizeof(double));
  if ((aux1d == NULL) || (aux2d == NULL))
    {
      if (p->incx != NULL) free(p->incx);
      if (aux1d != NULL) fftwf_free(aux1d);
      if (aux2d != NULL) fftwf_free(aux2d);
      free(p);
      return NULL;
    }
  fftwf_dft_c2r_x(init,p->d,p->x,p->incx,p->incmx[0],
		  p->y,p->incy,p->incmy[0],
		  p->n,p->m[0],p->isign,p->scale,
		  aux1d,&p->naux1,aux2d,&p->naux2);
  fftwf_free(aux1d);
  fftwf_free(aux2d);
  errstr(&ierno,s2015);
  /* allocate aux1 and aux2 arrays                            */
  p->aux1 = (double *) fftwf_malloc((_ESVINT64)p->naux1*sizeof(double));
  p->aux2 = (double *) fftwf_malloc((_ESVINT64)p->naux2*sizeof(double));
  if ((p->aux1 == NULL) || (p->aux2 == NULL))
    {
      if (p->incx != NULL) free(p->incx);
      if (p->aux1 != NULL) fftwf_free(p->aux1);
      if (p->aux2 != NULL) fftwf_free(p->aux2);
      free(p);
      return NULL;
    }
  /* initialization of fftwf_dft_c2r_x                        */
  fftwf_dft_c2r_x(init,p->d,p->x,p->incx,p->incmx[0],
		  p->y,p->incy,p->incmy[0],
		  p->n,p->m[0],p->isign,p->scale,
		  p->aux1,&p->naux1,p->aux2,&p->naux2);
  return p; 
}

fftwf_plan fftwf_plan_guru_dft_c2r(int rank, const fftwf_iodim *dims,
				 int howmany_rank, 
				 const fftwf_iodim *howmany_dims,
				 fftwf_complex *in, float *out,
				 unsigned flags)
{
  fftwf_plan p;
  double *aux1d, *aux2d;
  int (*iusadr) (); 
  int ierno,inoal,inomes,itrace,irange,dummy;
  char s2015[8];
  int size;
  int d, i;
  int init, stride;
  init = 1;
  if ((flags & FFTW_NO_SIMD) == FFTW_NO_SIMD) init = 2;
  if ((flags & FFTW_UNALIGNED) == FFTW_UNALIGNED) init = 2;
  if (rank > 1)
    {if ((flags & FFTW_PRESERVE_INPUT) == FFTW_PRESERVE_INPUT) return NULL;}
  /* allocate the structure fftwf_esv_plan_s                  */
  p = (fftwf_plan) malloc(sizeof(struct fftwf_esv_plan_s)); 
  if (p == NULL) return NULL;
  /* fill the structure fftwf_esv_plan_s                      */
  /* type of the transform: complex-to-complex,
     complex-to-real, or real-to-complex                      */
  p->type = ESV_C2R;
  /* dimension of the transform                               */
  p->d = rank;
  /* input array                                              */
  p->x = in;
  /* output array                                             */
  p->y = out;
  /* scale factor                                             */
  p->scale = 1.0;
  /* allocate all the integer arrays at the same time         */
  if (howmany_rank > 1)
    {
      size = 4*(rank + howmany_rank);
    }
  else
    {
      size = 4*rank + 3;
    }
  p->incx = (int *) malloc((_ESVINT64)size*sizeof(int));
  if (p->incx == NULL)
    {
      free(p);
      return NULL;
    }
  /* strides for the input array                              */
  /* strides for the output array                             */
  /* length of the transform                                  */
  /* direction of the transform                               */
  p->incy = &(p->incx[rank]);
  p->n = &(p->incy[rank]);
  p->isign = &(p->n[rank]);
  for (i=0; i<rank; i++)
    {
      p->incx[i] = dims[rank-1-i].is;
      p->incy[i] = dims[rank-1-i].os;
      p->n[i] = dims[rank-1-i].n;
      p->isign[i] = -1;
    }
  /* p->incmx, p->incmy, p->m, p->howmany_rank, and p->mm
     depend on the parameter howmany_rank                     */
  if (howmany_rank == 0)
    {
      p->howmany_rank = 1;
      p->mm = NULL;
      p->m = &(p->isign[rank]);
      p->incmx = &(p->m[1]);
      p->incmy = &(p->incmx[1]);

      if (rank == 0)
	{
	  p->m[0] = 1;
	  p->incmx[0] = 1;
	  if ((void*) in == (void*) out)
	    {
	      p->incmy[0] = 2;
	    }
	  else
	    {
	      p->incmy[0] = 1;
	    }
	}
      else
	{
	  p->m[0] = 1;
	  if (rank > 1)
	    {
	      p->incmx[0] = p->n[rank-1]*p->incx[rank-1];
	      p->incmy[0] = p->n[rank-1]*p->incy[rank-1];
	    }
	  else
	    {
	      p->incmx[0] = (p->n[0]/2+1)*p->incx[0];
	      if ((void*) in == (void*) out)
		{
		  p->incmy[0] = 2*(p->n[0]/2+1)*p->incy[0];
		}
	      else
		{
		  p->incmy[0] = p->n[0]*p->incy[0];
		}
	    }
	}
    }
  else if (howmany_rank == 1)
    {
      p->howmany_rank = 1;
      p->mm = NULL;
      p->m = &(p->isign[rank]);
      p->incmx = &(p->m[1]);
      p->incmy = &(p->incmx[1]);
 
      p->m[0] = howmany_dims[0].n;
      p->incmx[0] = howmany_dims[0].is;
      p->incmy[0] = howmany_dims[0].os;
    }
  else
    {
      p->howmany_rank = howmany_rank;
      p->m = &(p->isign[rank]);
      p->incmx = &(p->m[howmany_rank]);
      p->incmy = &(p->incmx[howmany_rank]);
      p->mm = &(p->incmy[howmany_rank]); 

      p->m[0] = howmany_dims[0].n;
      p->incmx[0] = howmany_dims[0].is;
      p->incmy[0] = howmany_dims[0].os;
      p->mm[0] = 1;
      for (i=1; i<howmany_rank; i++)
	{
	  p->m[i] = howmany_dims[i].n;
	  p->incmx[i] = howmany_dims[i].is;
	  p->incmy[i] = howmany_dims[i].os;

	  p->mm[i] = p->mm[0];
	  p->mm[0] = p->mm[0]*p->m[i];
	}

      /* here also we have to decide whether we will use      */
      /* SIMD performance enhancements                        */
      for (i=1; i<howmany_rank; i++)
	{
	  stride = p->incmx[i]/2;
	  if (p->incmx[i] != 2*stride) init = 2;
	  stride = p->incmy[i]/4;
	  if (p->incmy[i] != 4*stride) init = 2;
	}
    }


  /* size one dft                                             */
  d = 0;
  for (i=0; i<p->d; i++)
    {
      if (p->n[i] != 1)
	{d = d+1;}
    }

  if (d == 0) 
    {
      p->d = 0;
      return p;
    }

  /* determine size of aux1 and aux2 arrays                   */
  einfo(0,&dummy,&dummy);
  iusadr = enotrm;
  ierno = 2015;
  inoal = 0;
  inomes = -1;
  itrace = 0;
  irange = 2015;
  errsav(&ierno,s2015);
  errset(&ierno,&inoal,&inomes,&itrace,&iusadr,&irange);
  p->naux1 = 1+7*p->d;
  if (p->naux1 < 15) p->naux1=15;
  p->naux2 = 0;
  aux1d = (double *) fftwf_malloc((_ESVINT64)p->naux1*sizeof(double));
  aux2d = (double *) fftwf_malloc((_ESVINT64)2*sizeof(double));
  if ((aux1d == NULL) || (aux2d == NULL))
    {
      if (p->incx != NULL) free(p->incx);
      if (aux1d != NULL) fftwf_free(aux1d);
      if (aux2d != NULL) fftwf_free(aux2d);
      free(p);
      return NULL;
    }
  fftwf_dft_c2r_x(init,p->d,p->x,p->incx,p->incmx[0],
		  p->y,p->incy,p->incmy[0],
		  p->n,p->m[0],p->isign,p->scale,
		  aux1d,&p->naux1,aux2d,&p->naux2);
  fftwf_free(aux1d);
  fftwf_free(aux2d);
  errstr(&ierno,s2015);
  /* allocate aux1 and aux2 arrays                            */
  p->aux1 = (double *) fftwf_malloc((_ESVINT64)p->naux1*sizeof(double));
  p->aux2 = (double *) fftwf_malloc((_ESVINT64)p->naux2*sizeof(double));
  if ((p->aux1 == NULL) || (p->aux2 == NULL))
    {
      if (p->incx != NULL) free(p->incx);
      if (p->aux1 != NULL) fftwf_free(p->aux1);
      if (p->aux2 != NULL) fftwf_free(p->aux2);
      free(p);
      return NULL;
    }
  /* initialization of fftwf_dft_c2r_x                        */
  fftwf_dft_c2r_x(init,p->d,p->x,p->incx,p->incmx[0],
		 p->y,p->incy,p->incmy[0],
		 p->n,p->m[0],p->isign,p->scale,
		 p->aux1,&p->naux1,p->aux2,&p->naux2);
  return p; 
}

/* EXECUTE  ------------------------------------------------------ */


void fftwf_execute(fftwf_plan p)
{
  if (p->type == ESV_C2C)
    {
      fftwf_execute_dft(p,p->x,p->y);
    }
  else if (p->type == ESV_R2C)
    {
      fftwf_execute_dft_r2c(p,p->x,p->y);
    }
  else if (p->type == ESV_C2R)
    {
      fftwf_execute_dft_c2r(p,p->x,p->y);
    }

  return;
}

void fftwf_execute_dft(const fftwf_plan p,
		      fftwf_complex *in, fftwf_complex *out)
{
  fftwf_complex *x, *y;
  int lx, ly, rest, i, ii;
  int j[p->howmany_rank];

  x = in;
  y = out;

  if (p->howmany_rank == 1)
    {
      if (p->d == 0)
	{
	  fftwf_dft_size_one(x,y,p->m[0],p->incmx[0],p->incmy[0]);
	}
      else
	{
	  scftd(0,p->d,x,p->incx,p->incmx[0],
		y,p->incy,p->incmy[0],
		p->n,p->m[0],p->isign,p->scale,
		p->aux1,&p->naux1,p->aux2,&p->naux2);
	}
    }
  else
    {
      for (i=0; i<p->mm[0]; i++)
	{
	  rest = i;
	  for (ii=p->howmany_rank-1; ii>0; ii--)
	    {
	      j[ii] = rest/p->mm[ii];
	      rest = rest - j[ii]*p->mm[ii];
	    }
	  lx = j[1]*p->incmx[1];
	  ly = j[1]*p->incmy[1];
	  for (ii=2; ii<p->howmany_rank; ii++)
	    {
	      lx = lx + j[ii]*p->incmx[ii];
	      ly = ly + j[ii]*p->incmy[ii];
	    }
	  
	  if (p->d == 0)
	    {
	      fftwf_dft_size_one(&x[lx],&y[ly],
				p->m[0],p->incmx[0],p->incmy[0]);
	    }
	  else
	    {
	      scftd(0,p->d,&x[lx],p->incx,p->incmx[0],
		    &y[ly],p->incy,p->incmy[0],
		    p->n,p->m[0],p->isign,p->scale,
		    p->aux1,&p->naux1,p->aux2,&p->naux2);
	    }  
	}
    }
  return;
}

void fftwf_execute_dft_r2c(const fftwf_plan p,
			  float *in, fftwf_complex *out)
{
  float *x;
  fftwf_complex *y;
  int lx, ly, rest, i, ii;
  int j[p->howmany_rank];

  x = in;
  y = out;

  if (p->howmany_rank == 1)
    {
      if (p->d == 0)
	{
	  fftwf_dft_r2c_size_one(x,y,p->m[0],p->incmx[0],p->incmy[0]);
	}
      else
	{
	  srcftd(0,p->d,x,p->incx,p->incmx[0],
		 y,p->incy,p->incmy[0],
		 p->n,p->m[0],p->isign,p->scale,
		 p->aux1,&p->naux1,p->aux2,&p->naux2);
	}
    }
  else
    {
      for (i=0; i<p->mm[0]; i++)
	{
	  rest = i;
	  for (ii=p->howmany_rank-1; ii>0; ii--)
	    {
	      j[ii] = rest/p->mm[ii];
	      rest = rest - j[ii]*p->mm[ii];
	    }
	  lx = j[1]*p->incmx[1];
	  ly = j[1]*p->incmy[1];
	  for (ii=2; ii<p->howmany_rank; ii++)
	    {
	      lx = lx + j[ii]*p->incmx[ii];
	      ly = ly + j[ii]*p->incmy[ii];
	    }
	  
	  if (p->d == 0)
	    {
	      fftwf_dft_r2c_size_one(&x[lx],&y[ly],p->m[0],
				    p->incmx[0],p->incmy[0]);
	    }
	  else
	    {
	      srcftd(0,p->d,&x[lx],p->incx,p->incmx[0],
		     &y[ly],p->incy,p->incmy[0],
		     p->n,p->m[0],p->isign,p->scale,
		     p->aux1,&p->naux1,p->aux2,&p->naux2);
	    }  
	}
    }
  return;
}

void fftwf_execute_dft_c2r(const fftwf_plan p,
			  fftwf_complex *in, float *out)
{
  fftwf_complex *x;
  float *y;
  int lx, ly, rest, i, ii;
  int j[p->howmany_rank];

  x = in;
  y = out;

  if (p->howmany_rank == 1)
    {

      if (p->d == 0)
	{
	  fftwf_dft_c2r_size_one(x,y,p->m[0],p->incmx[0],p->incmy[0]);
	}
      else
	{
	  fftwf_dft_c2r_x(0,p->d,x,p->incx,p->incmx[0],
			  y,p->incy,p->incmy[0],
			  p->n,p->m[0],p->isign,p->scale,
			  p->aux1,&p->naux1,p->aux2,&p->naux2);
	}
    }
  else
    {
      for (i=0; i<p->mm[0]; i++)
	{
	  rest = i;
	  for (ii=p->howmany_rank-1; ii>0; ii--)
	    {
	      j[ii] = rest/p->mm[ii];
	      rest = rest - j[ii]*p->mm[ii];
	    }
	  lx = j[1]*p->incmx[1];
	  ly = j[1]*p->incmy[1];
	  for (ii=2; ii<p->howmany_rank; ii++)
	    {
	      lx = lx + j[ii]*p->incmx[ii];
	      ly = ly + j[ii]*p->incmy[ii];
	    }
	  
	  if (p->d == 0)
	    {
	      fftwf_dft_c2r_size_one(&x[lx],&y[ly],p->m[0],
				    p->incmx[0],p->incmy[0]);
	    }
	  else
	    {
	      fftwf_dft_c2r_x(0,p->d,&x[lx],p->incx,p->incmx[0],
			      &y[ly],p->incy,p->incmy[0],
			      p->n,p->m[0],p->isign,p->scale,
			      p->aux1,&p->naux1,p->aux2,&p->naux2);
	    }  
	}
    }
  return;
}


/* DESTROY  ------------------------------------------------------ */


void fftwf_destroy_plan(fftwf_plan p)
{
  /* first deallocate arrays in the structure fftwf_esv_plan_s    */
  free(p->incx);
  if (p->d != 0)
    {
      fftwf_free(p->aux1);
      fftwf_free(p->aux2);
    }
  /* now deallocate the structure fftwf_esv_plan_s               */
  free(p);
  return;
}

void fftwf_cleanup(void)
{
  return;
}


/* MEMORY  ------------------------------------------------------- */


void *fftwf_malloc(size_t size)
{
  void *p;
  size_t size_over_16;
  size_t pm_size;
  size_over_16 = size/16;
  pm_size = 16*size_over_16;
  if (pm_size < size) pm_size = pm_size + 16;
  if (posix_memalign(&p,16,pm_size) != 0) return NULL;
  return p;
}

void fftwf_free(void *p)
{
  free(p);
  return;
}


/* auxiliary kernelnels for size one ------------------------------ */


void fftwf_dft_size_one(fftwf_complex *x, fftwf_complex *y, 
		       int m, int incmx, int incmy)
{
  int lx, ly, i;
  lx = 0;
  ly = 0;
  for (i=0; i<m; i++)
    {
      y[ly][0] = x[lx][0];
      y[ly][1] = x[lx][1];
      lx = lx + incmx;
      ly = ly + incmy;
    }
  return;
}

void fftwf_dft_r2c_size_one(float *x, fftwf_complex *y, 
			   int m, int incmx, int incmy)
{
  int lx, ly, i;
  lx = 0;
  ly = 0;
  for (i=0; i<m; i++)
    {
      y[ly][0] = x[lx];
      y[ly][1] = 0.0;
      lx = lx + incmx;
      ly = ly + incmy;
    }
  return;
}

void fftwf_dft_c2r_size_one(fftwf_complex *x, float *y, 
			   int m, int incmx, int incmy)
{
  int lx, ly, i;
  lx = 0;
  ly = 0;
  for (i=0; i<m; i++)
    {
      y[ly] = x[lx][0];
      lx = lx + incmx;
      ly = ly + incmy;
    }
  return;
}

void fftwf_dft_c2r_x(int init, int d, void *x, int *incx, int incmx,
		     void *y, int *incy, int incmy, int *n, int m, 
		     int *isign, float scale, double *aux1, int *naux1, 
		     double *aux2, int *naux2)
{
  int (*iusadr) (); 
  int ierno,inoal,inomes,itrace,irange,dummy;
  char s2015[8];
  float *xx, *yy;
  float one;
  int naux1cc,naux2cc,naux1cr, naux2cr, naux1t, naux2t;
  int *pnaux1cc, *pnaux1cr;
  int initcc, initcr, a1cc, a1cr;
  int dcc, dcr, mcc;
  int lx, ly, lxx, lyy;
  int i, k, ii, nx, rest, nn[d], j[d];

  if (d == 1)
    {
      scrftd(init,d,x,incx,incmx,y,incy,incmy,
	     n,m,isign,scale,aux1,naux1,aux2,naux2);    
      return;
    }

  xx = (float*) x;
  yy = (float*) y;
  dcc = d-1;
  dcr = 1;
  mcc = n[0]/2+1;
  one = 1.0;

  if (init != 0)
    {
      /* initialization  */
      initcc = init;
      initcr = init;
      if ((m>1) && (incmx!=2*(incmx/2))) 
	{
	  initcc=2;
	  initcr=2;
	}
      for (ii=2; ii<d; ii++)
	{
	  if ((n[ii]>1) && (incx[ii]!=2*(incx[ii]/2))) 
	    {initcr=2;}
	}
      if ((m>1) && (incmy!=4*(incmy/4))) 
	{initcr=2;}
      for (ii=2; ii<d; ii++)
	{
	  if ((n[ii]>1) && (incy[ii]!=4*(incy[ii]/4))) 
	    {initcr=2;}
	}

      einfo(0,&dummy,&dummy);
      iusadr = enotrm;
      ierno = 2015;
      inoal = 0;
      inomes = -1;
      itrace = 0;
      irange = 2015;
      errsav(&ierno,s2015);
      errset(&ierno,&inoal,&inomes,&itrace,&iusadr,&irange);
      naux1cc = 7*d+1;
      naux2cc = 0;
      scftd(initcc,dcc,xx,&incx[1],incx[0],xx,&incx[1],incx[0],
	    &n[1],mcc,&isign[1],one,aux1,&naux1cc,aux2,&naux2cc);
      naux1cr = 15;
      naux2cr = 0;
      scrftd(initcr,dcr,xx,&incx[0],incx[1],yy,&incy[0],incy[1],
	     &n[0],n[1],&isign[0],scale,aux1,&naux1cr,aux2,&naux2cr);   
      errstr(&ierno,s2015);
      naux1t = naux1cc + naux1cr + 3;
      naux2t = naux2cc;
      if (naux2t < naux2cr) naux2t = naux2cr;

      if ((*naux1<naux1t) || (*naux2<naux2t))
	{
	  *naux1 = naux1t;
	  *naux2 = naux2t;
	  return;
	}

      a1cc = 2;
      a1cr = a1cc + naux1cc;
      if (a1cr != 2*(a1cr/2)) a1cr = a1cr+1;
      scftd(initcc,dcc,xx,&incx[1],incx[0],xx,&incx[1],incx[0],
	    &n[1],mcc,&isign[1],one,&aux1[a1cc],&naux1cc,aux2,naux2);
      scrftd(initcr,dcr,xx,&incx[0],incx[1],yy,&incy[0],incy[1],
	     &n[0],n[1],&isign[0],scale,&aux1[a1cr],&naux1cr,aux2,naux2); 

      pnaux1cc = (int*)&aux1[0];
      pnaux1cr = (int*)&aux1[1];
      *pnaux1cc = naux1cc;
      *pnaux1cr = naux1cr;
    }
  else
    {
      /* computation  */
      pnaux1cc = (int*)&aux1[0];
      pnaux1cr = (int*)&aux1[1];
      naux1cc = *pnaux1cc;
      naux1cr = *pnaux1cr;
      a1cc = 2;
      a1cr = a1cc + naux1cc;
      if (a1cr != 2*(a1cr/2)) a1cr = a1cr+1;

      nx = 1;
      for ( ii=2; ii<d; ii++)
	{
	  nn[ii]=nx;
	  nx=nx*n[ii];
	}
      lx = 0;
      ly = 0;
      for ( k=0; k<m; k++)
	{
	  scftd(init,dcc,&xx[lx],&incx[1],incx[0],&xx[lx],&incx[1],incx[0],
		&n[1],mcc,&isign[1],one,&aux1[a1cc],&naux1cc,aux2,naux2);
	  for ( i=0; i<nx; i++)
	    {
	      rest = i;
	      for ( ii=d-1; ii>1; ii--)
		{
		  j[ii] = rest/nn[ii];
		  rest = rest - j[ii]*nn[ii];
		}
	      lxx = lx;
	      lyy = ly;
	      for ( ii=2; ii<d; ii++)
		{
		  lxx = lxx + 2*j[ii]*incx[ii];
		  lyy = lyy + j[ii]*incy[ii];
		}
	      scrftd(init,dcr,&xx[lxx],&incx[0],incx[1],
		     &yy[lyy],&incy[0],incy[1],&n[0],n[1],
		     &isign[0],scale,&aux1[a1cr],&naux1cr,aux2,naux2); 
	    }
	  lx = lx + 2*incmx;
	  ly = ly + incmy;
	}
    }
}
