/**********************************************************************
*                                                                     *
*    MODULE-NAME = sfftw.c                                            *
*                                                                     *
*    DESCRIPTIVE-NAME = FFTW Fortran Interfaces (single precision).   *
*                                                                     *
*    Licensed Materials - Property of IBM                             *
*                                                                     *
*    5765-F82                                                         *
*    5765-G17                                                         *
*    5765-H25                                                         *
*    5765-L51                                                         *
*                                                                     *
*    Copyright IBM Corp. 2008, 2010                                   *
*    Copyright (c) 2003, 2007 Massachusetts Institute of Technology   *
*    Copyright (c) 2003, 2007 Matteo Frigo                            *
*    All Rights Reserved                                              *
*                                                                     *
*    US Government Users Restricted Rights - Use, duplication or      *
*    disclosure restricted by GSA ADP Schedule Contract with IBM Corp.*
*                                                                     *
*    FUNCTION = Call FFTW routines from Fortran                       *
*               for single precision computation.                     *
*                                                                     *
*    NOTES =                                                          *
*                                                                     *
*      DEPENDENCIES = (1) fftw3_essl.h                                *
*                                                                     *
*      RESTRICTIONS = None                                            *
*                                                                     *
*      PROCESSOR = C/C++ Compiler                                     *
*                                                                     *
***********************************************************************
*                                                                     *
*      PURPOSE = SEE FUNCTION                                         *
*                                                                     *
*      LINKAGE = NONE                                                 *
*                                                                     *
*      CONTROL BLOCKS = NONE                                          *
*                                                                     *
*      TABLES = NONE                                                  *
*                                                                     *
*      MACROS = NONE                                                  *
*                                                                     *
*      CHANGE-ACTIVITY = NONE                                         *
*                                                                     *
***********************************************************************
*                ALGORITHM DESCRIPTION                                *
*                                                                     *
* REFERENCES  = NONE                                                  *
*                                                                     *
* FUNCTION    =   See Above                                           *
*                                                                     *
***********************************************************************
*                    ALGORITHM EXAMPLES                               *
*                                                                     *
*   NONE                                                              *
*                                                                     *
***********************************************************************
*                                                                     *
*                PROCEDURAL LOGIC SECTION                             *
*                                                                     *
*     1.  FFTW Fortran interface routines are C functions that        *
*         will be called from Fortran code.                           *
*                                                                     *
*     2. FFTW Fortran interface routines are performed by calling     *
*        the FFTW C interface functions.                              *
*                                                                     *
*     3. FFTW Fortran interface routines expect multi dimensional     *
*        arrays to be in column-major order. Therefore FFTW Fortran   *
*        interface routines reverse the order of arrays before        *
*        calling FFTW C interface functions.                          *
*                                                                     *
*     4. The fftwf_iodim structure from the C guru interface splits   *
*        into separate arguments  (three integer array arguments) in  *
*        the Fortran guru interface.                                  * 
*                                                                     *
**********************************************************************/

#include "fftw3_essl.h"

void sfftw_plan_dft_1d(fftwf_plan *p, int *n, fftwf_complex *in, 
		       fftwf_complex *out, int *sign,unsigned *flags)
{
  *p = fftwf_plan_dft_1d(*n,in,out,*sign,*flags);
  return;
}

void sfftw_plan_dft_2d(fftwf_plan *p, int *nx, int *ny, 
		       fftwf_complex *in, fftwf_complex *out, 
		       int *sign,unsigned *flags)
{
  *p = fftwf_plan_dft_2d(*ny,*nx,in,out,*sign,*flags);
  return;
}

void sfftw_plan_dft_3d(fftwf_plan *p, int *nx, int *ny, int *nz,
		       fftwf_complex *in, fftwf_complex *out, 
		       int *sign,unsigned *flags)
{
  *p = fftwf_plan_dft_3d(*nz,*ny,*nx,in,out,*sign,*flags);
  return;
}

void sfftw_plan_dft(fftwf_plan *p, int *rank, int *n,
		    fftwf_complex *in, fftwf_complex *out, 
		    int *sign,unsigned *flags)
{
  int cn[*rank];
  for (int i=0; i<*rank; i++) cn[i] = n[*rank-i-1];
  *p = fftwf_plan_dft(*rank,cn,in,out,*sign,*flags);
  return;
}

void sfftw_plan_many_dft(fftwf_plan *p, 
			 int *rank, int *n, int *howmany,
			 fftwf_complex *in, int *inembed,
			 int *istride, int *idist,
			 fftwf_complex *out, int *onembed,
			 int *ostride, int *odist,  
			 int *sign,unsigned *flags)
{
  int j;
  int cn[*rank], cinembed[*rank], conembed[*rank];
  
  j = *rank-1;
  
  if ((inembed != NULL) &&  (onembed != NULL))
    {
      for (int i=0; i<*rank; i++)
	{
	  cn[i] = n[j];
	  cinembed[i] = inembed[j];
	  conembed[i] = onembed[j];
	  j--;
	}
      *p = fftwf_plan_many_dft(*rank,cn,*howmany,in,cinembed,*istride, *idist,
			       out,conembed,*ostride,*odist,*sign,*flags);
    }
  else if ((inembed == NULL) &&  (onembed != NULL))
    {
      for (int i=0; i<*rank; i++)
	{
	  cn[i] = n[j];
	  conembed[i] = onembed[j];
	  j--;
	}
      *p = fftwf_plan_many_dft(*rank,cn,*howmany,in,NULL,*istride, *idist,
			       out,conembed,*ostride,*odist,*sign,*flags);
    }
  else if ((inembed != NULL) &&  (onembed == NULL))
    {
      for (int i=0; i<*rank; i++)
	{
	  cn[i] = n[j];
	  cinembed[i] = inembed[j];
	  j--;
	}
      *p = fftwf_plan_many_dft(*rank,cn,*howmany,in,cinembed,*istride, *idist,
			       out,NULL,*ostride,*odist,*sign,*flags);
    }
  else 
    {
      for (int i=0; i<*rank; i++)
	{
	  cn[i] = n[j];
	  j--;
	}
      *p = fftwf_plan_many_dft(*rank,cn,*howmany,in,NULL,*istride, *idist,
			       out,NULL,*ostride,*odist,*sign,*flags);
    }
  
  return;
}

void sfftw_plan_guru_dft(fftwf_plan *p, int *rank, int *dims_n,
			 int *dims_is, int *dims_os, 
			 int *howmany_rank, int *howmany_dims_n,
			 int *howmany_dims_is, int *howmany_dims_os, 
			 fftwf_complex *in, fftwf_complex *out,
			 int *sign,unsigned *flags)
{
  int j;
  fftwf_iodim dims[*rank], howmany_dims[*howmany_rank];
  
  j = *rank-1;
  for (int i=0; i<*rank; i++)
    {
      dims[i].n = dims_n[j];
      dims[i].is = dims_is[j];
      dims[i].os = dims_os[j];
      j--;
    }
  
  for (int i=0; i<*howmany_rank; i++)
    {
      howmany_dims[i].n = howmany_dims_n[i];
      howmany_dims[i].is = howmany_dims_is[i];
      howmany_dims[i].os = howmany_dims_os[i];
    }
  
  *p = fftwf_plan_guru_dft(*rank, dims, *howmany_rank, howmany_dims,
			   in, out, *sign, *flags);
  
  return;
}

void sfftw_plan_dft_r2c_1d(fftwf_plan *p, int *n, float *in, 
			   fftwf_complex *out, unsigned *flags)
{
  *p = fftwf_plan_dft_r2c_1d(*n,in,out,*flags);
  return;
}

void sfftw_plan_dft_r2c_2d(fftwf_plan *p, int *nx, int *ny, 
			   float *in, fftwf_complex *out, 
			   unsigned *flags)
{
  *p = fftwf_plan_dft_r2c_2d(*ny,*nx,in,out,*flags);
  return;
}

void sfftw_plan_dft_r2c_3d(fftwf_plan *p, int *nx, int *ny, int *nz, 
			   float *in, fftwf_complex *out, 
			   unsigned *flags)
{
  *p = fftwf_plan_dft_r2c_3d(*nz,*ny,*nx,in,out,*flags);
  return;
}

void sfftw_plan_dft_r2c(fftwf_plan *p, int *rank, int *n,
			float *in, fftwf_complex *out, 
			unsigned *flags)
{
  int cn[*rank];
  for (int i=0; i<*rank; i++) cn[i] = n[*rank-i-1];
  *p = fftwf_plan_dft_r2c(*rank,cn,in,out,*flags);
  return;
}

void sfftw_plan_many_dft_r2c(fftwf_plan *p,
			     int *rank, int *n, int *howmany,
			     float *in, int *inembed,
			     int *istride, int *idist,
			     fftwf_complex *out, int *onembed,
			     int *ostride, int *odist,  
			     unsigned *flags)
{
  int j;
  int cn[*rank], cinembed[*rank], conembed[*rank];
  
  j = *rank-1;
  
  if ((inembed != NULL) &&  (onembed != NULL))
    {
      for (int i=0; i<*rank; i++)
	{
	  cn[i] = n[j];
	  cinembed[i] = inembed[j];
	  conembed[i] = onembed[j];
	  j--;
	}
      *p = fftwf_plan_many_dft_r2c(*rank,cn,*howmany,
				   in,cinembed,*istride, *idist,
				   out,conembed,*ostride,*odist,*flags);
    }
  else if ((inembed == NULL) &&  (onembed != NULL))
    {
      for (int i=0; i<*rank; i++)
	{
	  cn[i] = n[j];
	  conembed[i] = onembed[j];
	  j--;
	}
      *p = fftwf_plan_many_dft_r2c(*rank,cn,*howmany,
				   in,NULL,*istride, *idist,
				   out,conembed,*ostride,*odist,*flags);
    }
  else if ((inembed != NULL) &&  (onembed == NULL))
    {
      for (int i=0; i<*rank; i++)
	{
	  cn[i] = n[j];
	  cinembed[i] = inembed[j];
	  j--;
	}
      *p = fftwf_plan_many_dft_r2c(*rank,cn,*howmany,
				   in,cinembed,*istride, *idist,
				   out,NULL,*ostride,*odist,*flags);
    }
  else 
    {
      for (int i=0; i<*rank; i++)
	{
	  cn[i] = n[j];
	  j--;
	}
      *p = fftwf_plan_many_dft_r2c(*rank,cn,*howmany,
				   in,NULL,*istride, *idist,
				   out,NULL,*ostride,*odist,*flags);
    }
  
  return;
}

void sfftw_plan_guru_dft_r2c(fftwf_plan *p, int *rank, int *dims_n,
			     int *dims_is, int *dims_os,
			     int *howmany_rank, int *howmany_dims_n,
			     int *howmany_dims_is, int *howmany_dims_os,
			     float *in, fftwf_complex *out,
			     unsigned *flags)
{
  int j;
  fftwf_iodim dims[*rank], howmany_dims[*howmany_rank];
  
  j = *rank-1;
  for (int i=0; i<*rank; i++)
    {
      dims[i].n = dims_n[j];
      dims[i].is = dims_is[j];
      dims[i].os = dims_os[j];
      j--;
    }
  
  for (int i=0; i<*howmany_rank; i++)
    {
      howmany_dims[i].n = howmany_dims_n[i];
      howmany_dims[i].is = howmany_dims_is[i];
      howmany_dims[i].os = howmany_dims_os[i];
    }
  
  *p = fftwf_plan_guru_dft_r2c(*rank, dims, *howmany_rank, howmany_dims,
			       in, out, *flags);
  
  return;
}

void sfftw_plan_dft_c2r_1d(fftwf_plan *p, int *n, fftwf_complex *in, 
			   float *out, unsigned *flags)
{
  *p = fftwf_plan_dft_c2r_1d(*n,in,out,*flags);
  return;
}

void sfftw_plan_dft_c2r_2d(fftwf_plan *p, int *nx, int *ny, 
			   fftwf_complex *in, float *out, 
			   unsigned *flags)
{
  *p = fftwf_plan_dft_c2r_2d(*ny,*nx,in,out,*flags);
  return;
}

void sfftw_plan_dft_c2r_3d(fftwf_plan *p, int *nx, int *ny, int *nz, 
			   fftwf_complex *in, float *out, 
			   unsigned *flags)
{
  *p = fftwf_plan_dft_c2r_3d(*nz,*ny,*nx,in,out,*flags);
  return;
}

void sfftw_plan_dft_c2r(fftwf_plan *p, int *rank, int *n,
			fftwf_complex *in, float *out, 
			unsigned *flags)
{
  int cn[*rank];
  for (int i=0; i<*rank; i++) cn[i] = n[*rank-i-1];
  *p = fftwf_plan_dft_c2r(*rank,cn,in,out,*flags);
  return;
}

void sfftw_plan_many_dft_c2r(fftwf_plan *p,  
			     int *rank, int *n, int *howmany,
			     fftwf_complex *in, int *inembed,
			     int *istride, int *idist,
			     float *out, int *onembed,
			     int *ostride, int *odist,  
			     unsigned *flags)
{
  int j;
  int cn[*rank], cinembed[*rank], conembed[*rank];
  
  j = *rank-1;
  
  if ((inembed != NULL) &&  (onembed != NULL))
    {
      for (int i=0; i<*rank; i++)
	{
	  cn[i] = n[j];
	  cinembed[i] = inembed[j];
	  conembed[i] = onembed[j];
	  j--;
	}
      *p = fftwf_plan_many_dft_c2r(*rank,cn,*howmany,
				   in,cinembed,*istride, *idist,
				   out,conembed,*ostride,*odist,*flags);
    }
  else if ((inembed == NULL) &&  (onembed != NULL))
    {
      for (int i=0; i<*rank; i++)
	{
	  cn[i] = n[j];
	  conembed[i] = onembed[j];
	  j--;
	}
      *p = fftwf_plan_many_dft_c2r(*rank,cn,*howmany,
				   in,NULL,*istride, *idist,
				   out,conembed,*ostride,*odist,*flags);
    }
  else if ((inembed != NULL) &&  (onembed == NULL))
    {
      for (int i=0; i<*rank; i++)
	{
	  cn[i] = n[j];
	  cinembed[i] = inembed[j];
	  j--;
	}
      *p = fftwf_plan_many_dft_c2r(*rank,cn,*howmany,
				   in,cinembed,*istride, *idist,
				   out,NULL,*ostride,*odist,*flags);
    }
  else 
    {
      for (int i=0; i<*rank; i++)
	{
	  cn[i] = n[j];
	  j--;
	}
      *p = fftwf_plan_many_dft_c2r(*rank,cn,*howmany,
				   in,NULL,*istride, *idist,
				   out,NULL,*ostride,*odist,*flags);
    }
  
  return;
}

void sfftw_plan_guru_dft_c2r(fftwf_plan *p, int *rank, int *dims_n,
			     int *dims_is, int *dims_os,
			     int *howmany_rank, int *howmany_dims_n,
			     int *howmany_dims_is, int *howmany_dims_os,
			     fftwf_complex *in, float *out,
			     unsigned *flags)
{
  int j;
  fftwf_iodim dims[*rank], howmany_dims[*howmany_rank];
  
  j = *rank-1;
  for (int i=0; i<*rank; i++)
    {
      dims[i].n = dims_n[j];
      dims[i].is = dims_is[j];
      dims[i].os = dims_os[j];
      j--;
    }
  
  for (int i=0; i<*howmany_rank; i++)
    {
      howmany_dims[i].n = howmany_dims_n[i];
      howmany_dims[i].is = howmany_dims_is[i];
      howmany_dims[i].os = howmany_dims_os[i];
    }

  *p = fftwf_plan_guru_dft_c2r(*rank, dims, *howmany_rank, howmany_dims,
			       in, out, *flags);
  
  return;
}

void sfftw_execute(fftwf_plan *p)
{
  fftwf_execute(*p);
  return;
}

void sfftw_execute_dft(fftwf_plan *p,
		       fftwf_complex *in, fftwf_complex *out)
{
  fftwf_execute_dft(*p,in,out);
  return;
}

void sfftw_execute_dft_r2c(fftwf_plan *p,
			   float *in, fftwf_complex *out)
{
  fftwf_execute_dft_r2c(*p,in,out);
  return;
}

void sfftw_execute_dft_c2r(fftwf_plan *p,
			   fftwf_complex *in, float *out)
{
  fftwf_execute_dft_c2r(*p,in,out);
  return;
}

void sfftw_destroy_plan(fftwf_plan *p)
{
  fftwf_destroy_plan(*p);
  return;
}

void sfftw_cleanup(void)
{
  fftwf_cleanup();
  return;
}


